/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dbrent.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */
#include <stdio.h>
#include <math.h>
#include "tools.h"

#define ZEPS 1.e-20 /* effective machine accuracy */

/* One dimensional line minimization.  Upon return, the function
 * has been called at the minimum.
 * 
 * Input:
 *   ax,bx,cx = values below, near, above minimum
 *   fb,dfb = function values at bx.
 *   fun() = function to calculate function and derivative.
 *	It also has attached an `environment' pointer (type 
 *	void) that just gets passed to it.  See linmin() for 
 *	an example of why this is nice.
 *   tol = fractional tolerance on function value at minimum
 *   iter = maximum iterations
 *
 * Output:
 *   xmin = parameter minimizing the function
 *   fret = minimized function value
 *
 *   return value = 0 if all went as planned, -1 for too many
 *   iterations, or a non-zero return code from fun() if some
 *   parameter between ax and cx was an invalid parameter.
 *
 * David Deaven deaven@ishmael.ameslab.gov, start w/numerical recipes.
 */
int dbrent(double ax, double bx, double cx, double fb, double dfb,
	int (*fun)(double x, double *y, double *df, void *fenv),
	void *fenv, double tol, int iter, double *xmin, double *fret)
{
    double olde, dw, fu, du, xm;
    double a, b, d, e, u, v, w, x, d1, d2, u1, u2, fv, fw, fx, dx, dv; 
    int ok1, ok2, r = 0;
    double tol1, tol2;

    a = ax < cx ? (b=cx, ax) : (b=ax, cx);

    x = w = v = bx;
    fx = fw = fv = fb;
    dx = dw = dv = dfb;

    e = d = 0.0;

    while(iter-- > 0)
    {
	xm = (a + b) * 0.5;
	tol1 = tol * fabs(x) + ZEPS;
	tol2 = tol1 * 2.0;
	if (fabs(x - xm) <= tol2 - (b - a) * 0.5) goto done;
	if (fabs(e) > tol1) {
	    d1 = (b - a) * 2.0;
	    d2 = d1;
	    if (dw != dx) { /* secant method, w point */
		d1 = (w - x) * dx / (dx - dw);
	    }
	    if (dv != dx) { /* secant method, v point */
		d2 = (v - x) * dx / (dx - dv);
	    }
	    u1 = x + d1;
	    u2 = x + d2;
	    ok1 = (a - u1) * (u1 - b) > 0.0 && dx * d1 <= 0.0;
	    ok2 = (a - u2) * (u2 - b) > 0.0 && dx * d2 <= 0.0;
	    olde = e;
	    e = d;

	    /* Only take an "acceptable" d.
	     * If both are acceptable, take the smaller.
	     */
	    if (ok1 || ok2)
	    {
		if (ok1 && ok2)
		  d = (fabs(d1) < fabs(d2) ? d1 : d2);
		else if (ok1)
		  d = d1;
		else
		  d = d2;
		
		if (fabs(d) <= fabs(olde * 0.5))
		{
		    u = x + d;
		    if ((u - a) < tol2 || (b - u) < tol2)
		      d = copysign(tol1, xm - x);
		    
		    goto skip;
		}
	    }
	}
	/* Decide which segment by the sign of the derivative,
	 * and perform bisection.
	 */
	e = (dx >= 0.0 ? a - x : b - x);
	d = e * 0.5;
skip:
	if (fabs(d) >= tol1)
	{
	    u = x + d;
	    if((r=(*fun)(u, &fu, &du, fenv)))return r;
	}
	else
	{
	    u = x + copysign(tol1, d);
	    if((r=(*fun)(u, &fu, &du, fenv)))return r;
	    if (fu > fx) goto done; /* smallest step downhill is uphill */
	}
	if (fu <= fx)
	{
	    if (u >= x)
	      a = x;
	    else
	      b = x;

	    v = w; fv = fw; dv = dw;
	    w = x; fw = fx; dw = dx;
	    x = u; fx = fu; dx = du;
	}
	else
	{
	    if (u < x)
	      a = u;
	    else
	      b = u;
	    
	    if (fu <= fw || w == x)
	    {
		v = w; fv = fw; dv = dw;
		w = u; fw = fu; dw = du;
	    }
	    else if (fu <= fv || v == x || v == w)
	    {
		v = u; fv = fu; dv = du;
	    }
	}
    } /* iter */
    r = -1;

done:
    *xmin = x;
    *fret = fx;
    return r;
}
