/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dmdman2html.c,v 1.6 1996/06/11 14:54:51 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "memory.h"

#define CHUNK 1024
#define isP(PC) ( \
     ( isprint(*(PC)) && !isspace(*(PC)) && !strchr("()", *(PC)) ) \
  || ( *(PC) == '\n' && *(PC - 1) == '-') \
  || ( *(PC) == '-' && *(PC + 1) == '\n') \
  )

static char *usage = "usage: %s [-title title] [-body options]\n";

/* replace ASCII characters with html escape sequences
 */
static void puthtmlchar(int c)
{
	switch(c) {
	  case '"': printf("&#22"); break;
	  case '&': printf("&#38"); break;
	  case '<': printf("&#60"); break;
	  case '>': printf("&#62"); break;
	  default: putchar(c); break;
	}
}

static int is_cr_space(char *start, char *s)
{
	while(s > (start + 2) && (*s == ' ' || *s == '\t')) s--;
	if(*s == '\n' && *(s - 1) == '-') return 1;
	return 0;
}

static char *readref(char *start, unsigned int offset, int *backspaces)
{
	int i, n;
	char *t, *s = start + offset;
	static char *buf = NULL;

	/* first, check the section number */
	if(*s != '(') return NULL;
	if(!strchr("135", *(s+1))) return NULL;
	if(*(s+2) != ')') return NULL;

	/* scan backwards, to find the start of the name */
	t = s;
	while(t > start) {
	  if(
	    (isprint(*(t-1)) && !isspace(*(t-1)) && !strchr("()", *(t-1))) ||
	    (is_cr_space(start, t - 1))
	  ) t--;
	  else break;
	}
	if(t == s) return NULL;

	/* copy the name into our local buffer */
	*backspaces = s - t;
	buf = myrealloc(buf, (s - t + 2) * sizeof(*buf));
	for(i=n=0; (t + i) < s; i++) {
	  buf[n++] = t[i];
	  if(t[i] == ' ' || t[i] == '\t') n -= 1;
	  else if(t[i] == '\n') n -= 2;
	}
	buf[n] = '\0';

	return buf;
}

static int match(char *s1, char *s2)
{
	return (strcmp(s1, s2) == 0 ? 1 : 0);
}

void main(int argc, char **argv)
{
	int arg, ic, backspaces;
	unsigned int i, j, len = 0, dlen = 0;
	char *buffer = NULL;
	char *refname;

	/* variables which may be modified by the command line */
	char *title = "manual page";
	char *body_options =
	  "bgcolor=\"550b613a6d4f\" text=\"white\""
	  "link=\"ad91ec36ffff\" vlink=\"0000c312ffff\""
	  ;

	/* process the command line arguments */
	for(arg=1; arg<argc; arg++) {
	  if(match(argv[arg], "-title"))
	  {
	    title = argv[++arg];
	  }
	  else if(match(argv[arg], "-body"))
	  {
	    body_options = argv[++arg];
	  }
	  else if(match(argv[arg], "-V") || match(argv[arg], "-help"))
	  {
	    fprintf(stderr, usage, argv[0]);
	    exit(0);
	  }
	  else
	  {
	    fprintf(stderr, usage, argv[0]);
	    exit(1);
	  }
	}

	/* Read the text into memory */
	while((ic = getchar()) != EOF) {
	  if(len + 1 > dlen) {
	    dlen += CHUNK;
	    buffer = myrealloc(buffer, (dlen + 1) * sizeof(*buffer));
	  }
	  buffer[len++] = ic;
	}

	/* print an HTML header */
	printf(
	  "<html>\n"
	  "<head>\n"
	  "<title>%s</title>\n"
	  "</head>\n"
	  "<body %s>\n"
	  "<h1>%s</h1>\n"
	  "<pre>\n", 
	  title,
	  body_options,
	  title
	);

	/* Output the text, substituting a hyperlink for
	 * patterns which match `P([135])' where P is a sequence
	 * of non-white printable characters.
	 */
	for(i=0; i<len;) {
	  /* find the next '('... */
	  for(j=i; j < len && buffer[j] != '('; j++);
	  if((refname = readref(buffer, j, &backspaces)))
	  {
	    while(i < (j - backspaces)) puthtmlchar(buffer[i++]);
	    /* output href text only for non-white characters in
	     * this sequence.
	     */
	    { char href[FILENAME_MAX], inref = 1;
	      sprintf(href, "<a href=\"%s.%c.html\">", refname, buffer[j + 1]);
	      printf("%s", href);
	      i = j - backspaces;
	      while(i < j) {
		if(isspace(buffer[i])) {
		  if(inref) {
		    printf("</a>"); 
		    inref = 0;
		  }
		} else {
		  if(!inref) {
		    printf("%s", href);
		    inref = 1;
		  }
		}
		puthtmlchar(buffer[i++]);
	      }
	      if(inref) printf("</a>");
	    }
	  }
	  while(i < len && i <= (j + 2)) puthtmlchar(buffer[i++]);
	}

	/* print the HTML footer
	 */
	printf(
	  "</pre>\n"
	  "</body>\n"
	  "</html>\n"
	);

	exit(0);
}
