/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dmdpp.c,v 1.12 1996/06/25 21:30:57 deaven Exp $
 */

/* This is a simple implementation of a greatly restricted set of cpp
 * functions.  The only ones supported are
 *
 * #include "file"
 *
 * #define name value
 *
 * #undef name
 *
 * #ifdef name  <or>  #ifndef name
 * ...
 * #else
 * ...
 * #endif
 *
 * #define name(arg1,...,argn) macro-text
 *
 * #error message
 *
 * This code is not very robust!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX_LINE_LENGTH 2048

#ifndef SYSTEM
#define SYSTEM "__unix__"
#endif

#if !defined(DMDDIR) && !defined(USR_TOPDIR)
 #define DMDDIR "."  	/* these should generate errors, right? */
 #define USR_TOPDIR "." 
#elif defined(DMDDIR) && !defined(USR_TOPDIR)
 #define USR_TOPDIR DMDDIR
#elif !defined(DMDDIR) && defined(USR_TOPDIR)
 #define DMDDIR USR_TOPDIR
#else
 #define BOTH_DMD_TOP_DEFINED
#endif /* defined(DMDDIR), defined(USR_TOPDIR) */

#ifndef FILENAME_MAX
#define FILENAME_MAX 512
#endif

static char *usage = "usage: %s [-Isearch-dir] [-Dsymbol[=value]] [file]\n";

struct pp_define {
	int nargs;	/* number of arguments (zero if not a macro) */
	int nchar;	/* chars in search */
	char *search;	/* search string or macro name */
	char *replace;	/* replace string or macro text */
	char **args;	/* argument names */
};

struct pp_if {
	int on;
};

static struct pp_define *defs = NULL;
static int ndefs = 0;

static struct pp_if *ifs = NULL;
static int nifs = 0;

static char **sdirs = NULL;
static int nsdirs = 0;

static char *myname = "dmdpp";

enum pp_op {
  PP_INCLUDE,
  PP_ERROR,
  PP_DEFINE,
  PP_UNDEF,
  PP_IFDEF,
  PP_IFNDEF,
  PP_ELSE,
  PP_ENDIF,
  PP_NOP
};

struct pp_operator {
	char *name;
	enum pp_op op;
};
static struct pp_operator pp_operators[] = {
  {"#include",	PP_INCLUDE},
  {"#error",	PP_ERROR},
  {"#define",	PP_DEFINE},
  {"#undef",	PP_UNDEF},
  {"#ifdef",	PP_IFDEF},
  {"#ifndef",	PP_IFNDEF},
  {"#else",	PP_ELSE},
  {"#endif",	PP_ENDIF},
  {"",		PP_NOP}
};

static void *getmem(int size)
{
	void *mem = malloc(size);
	if(mem == NULL) {
	  fprintf(stderr, "%s: error getting %d bytes of memory\n",
	    myname, size);
	  exit(1);
	}
	return mem;
}

static void *resizemem(void *mem, int size)
{
	if(mem == NULL) return getmem(size);

	mem = realloc(mem, size);
	if(mem == NULL) {
	  fprintf(stderr, "%s: error resizing %d bytes of memory\n",
	    myname, size);
	  exit(1);
	}
	return mem;
}

static char *stringcopy(char *s)
{
	char *r;

	r = getmem((strlen(s) + 1) * sizeof(char));
	strcpy(r, s);
	return r;
}

static int which_macro(char *s)
{
	int i;
	for(i=0; i<ndefs; ++i)
	  if(strcmp(defs[i].search, s) == 0) return i;
	return -1;
}

static struct pp_operator *which_pp_operator(char *s)
{
	struct pp_operator *op;
	for(op = pp_operators; op->op != PP_NOP; ++op)
	  if(strlen(s) >= strlen(op->name) &&
	     strncmp(s, op->name, strlen(op->name)) == 0) return op;
	return NULL;
}

static char **get_macro_args(char *first, char **last, int *nargs)
{
	int nest, done = 0;
	char *s, *e, **args = NULL;
	*nargs = 0;

	while(*first != '\0' && *first != '(') ++first;
	if(*first == '\0') return args;
	++first;
	nest = 0;

	for(s = first; *s != '\0' && *s != ')' ; ++s) {
	  for(e = s; *e != '\0' && *e != ',' &&
	      (nest != 0 || *e != ')'); ++e) {
	    if(*e == '(') ++nest;
	    if(*e == ')') --nest;
	  }
	  if(*e == '\0') goto bad_syntax;
	  if(*e == ')') done = 1;
	  *e = '\0';
	  args = resizemem(args, ((*nargs) + 1) * sizeof(char *));
	  args[*nargs] = stringcopy(s);
	  ++(*nargs);
	  s = e; 
	  if(done) break;
	}
	*last = e;
	return args;

bad_syntax:
	fprintf(stderr, "%s: error: bad macro argument syntax \"%s\"\n",
	  myname, first);
	exit(1);
}

static void add_def(char *search, char *replace)
{
	int n;
	char *s, *e;

	if((n = which_macro(search)) < 0) {
	  defs = resizemem(defs, (ndefs + 1) * sizeof(struct pp_define));
	  n = ndefs++;
	}

	if((s = strchr(search, '(')) != NULL) {
	  defs[n].args = get_macro_args(s, &e, &(defs[n].nargs));
	  *s = '\0';
	  while(isspace(*(s - 1))) *(--s) = '\0';
	}
	else {
	  defs[n].nargs = 0;
	  defs[n].args = NULL;
	}

	defs[n].nchar = strlen(search);
	defs[n].search = stringcopy(search);
	defs[n].replace = stringcopy(replace);
}

static char *expand_macro(char *first, char **last, struct pp_define *macro)
{
	char *s, *t, **args;
	int i, replaced, nargs;
	static char buf[MAX_LINE_LENGTH];
	
	if(macro->nargs == 0) {
	  *last = first - 1;
	  return macro->replace;
	}

	args = get_macro_args(first, last, &nargs);
	if(nargs != macro->nargs) {
	  fprintf(stderr, "%s: error: wrong number of macro args for %s\n",
	    myname, macro->search);
	  exit(1);
	}

	t = buf;
	s = macro->replace;
	while(*s != '\0') {
	  replaced = 0;
	  for(i=0; i<nargs; ++i)
	    if(strncmp(s, macro->args[i], strlen(macro->args[i])) == 0) {
	      strcpy(t, args[i]);
	      t += strlen(args[i]);
	      s += strlen(macro->args[i]);
	      replaced = 1;
	      break;
	    }
	  if(!replaced) *(t++) = *(s++);
	}
	*t = '\0';

	return buf;
}

static void push_if(int on)
{
	ifs = resizemem(ifs, (nifs + 1) * sizeof(struct pp_if));
	ifs[nifs++].on = on;
}

static void pop_if()
{
	if(nifs < 1) {
	  fprintf(stderr, "%s: error -- too many #endif\n", myname);
	  exit(1);
	}

	--nifs;
}

static int flameon()
{
	int i;
	for(i=0; i<nifs; ++i) if(!ifs[i].on) return 0;
	return 1;
}

static int perform_subs(char *s0)
{
	int i, replaced, n = 0;
	char *s = s0, *t, *e, *m;
	static char buf[MAX_LINE_LENGTH];

	t = buf;

	while(*s != '\0') {
	  replaced = 0;
	  for(i=0; i<ndefs; ++i)
	    if(strncmp(defs[i].search, s, defs[i].nchar) == 0) {
	      m = expand_macro(s + defs[i].nchar, &e, defs + i);
	      strcpy(t, m);
	      t += strlen(m);
	      s = e + 1;
	      replaced = 1;
	      ++n;
	      break;
	    }
	  if(!replaced) *(t++) = *(s++);
	}

	*t = '\0';
	strcpy(s0, buf);

	return n;
}

static void print_line(char *s)
{
	while(perform_subs(s));
	fputs(s, stdout);
}

static void remove_newline(char *s)
{
	int i = strlen(s);
	for(i = strlen(s) - 1; i >= 0 && s[i] == '\n'; --i) s[i] = '\0';
}

static FILE *sdiropen(char *fn)
{
	int i;
	char buf[FILENAME_MAX];
	FILE *f;

	for(i=0; i<nsdirs; ++i) {
	  sprintf(buf, "%s/%s", sdirs[i], fn);
	  f = fopen(buf, "r");
	  if(f != NULL) return f;
	}

	return NULL;
}

static void do_file(char *fn)
{
	int i, j, on, line = 0;
	char *s, *ifn, *search, *replace;
	char buf[MAX_LINE_LENGTH];
	FILE *f;
	struct pp_operator *op;

	if(strcmp(fn, "-") == 0)
	  f = stdin;
	else {
	  f = sdiropen(fn);
	  if(f == NULL) {
	    fprintf(stderr, "%s: cannot open \"%s\"\n", myname, fn);
	    exit(1);
	  }
	}

	while(fgets(buf, MAX_LINE_LENGTH, f)) {

	  ++line;

	  if((op = which_pp_operator(buf)) != NULL) {
	    remove_newline(buf);
	    while(buf[strlen(buf) - 1] == '\\') {
	      fgets(buf + strlen(buf) - 1, MAX_LINE_LENGTH - strlen(buf), f);
	      remove_newline(buf);
	      if(feof(f)) break;
	    }
	    for(s = buf + strlen(op->name); isspace(*s); ++s);
	    switch(op->op) {
	      case PP_INCLUDE:
		if(flameon()) {
		  ifn = s;
		  if(*ifn == '\"') {
		    ++ifn;
		    if(strlen((s = ifn)) < 2) goto bad_syntax;
		    while(*s != '\0' && *s != '\"') ++s;
		    if(*s == '\"')
		      *s = '\0';
		    else
		      goto bad_syntax;
		  }
		  do_file(ifn);
		}
		break;

	      case PP_ERROR:
		if(flameon()) {
		  fprintf(stderr, "%s: error -- %s\n", myname, s);
		  exit(1);
		}
		break;

	      case PP_DEFINE:
		if(flameon()) {
		  search = s;
		  while(*s != '\0' && !isspace(*s)) ++s;
		  if(*s == '\0')
		    replace = "";
		  else {
		    *(s++) = '\0';
		    while(*s != '\0' && isspace(*s)) ++s;
		    if(*s == '\0')
		      replace = "";
		    else
		      replace = s;
		  }
		  add_def(search, replace);
		}
		break;

	      case PP_UNDEF:
		if(flameon()) {
		  search = s;
		  while(*s != '\0' && !isspace(*s)) ++s;
		  *s = '\0';
		  i = which_macro(search);
		  if(i >= 0) {
		    for(j=i; j<(ndefs - 1); ++j) defs[j] = defs[j+1];
		    --ndefs;
		  }
		}
		break;

	      case PP_IFDEF:
	      case PP_IFNDEF:
		search = s;
		while(*s != '\0' && !isspace(*s)) ++s;
		*s = '\0';
		on = 0;
		for(i=0; i<ndefs; ++i) 
		  if(strcmp(search, defs[i].search) == 0) on = 1;
		push_if(op->op == PP_IFDEF ? on : !on);
		break;

	      case PP_ELSE:
		if(nifs < 1) goto bad_syntax;
		ifs[nifs - 1].on = !(ifs[nifs - 1].on);
		break;

	      case PP_ENDIF:
		pop_if();
		break;

	      case PP_NOP:
		break;

	    } /* switch(op->op) */
	  } /* if buf[] is an operator line */

	  else
	    if(flameon()) print_line(buf);
	}

	fclose(f);
	return;

bad_syntax:
	fprintf(stderr, "%s: syntax error, `%s' line %d: %s\n",
	  myname, fn, line, buf);
	exit(1);

}

static void add_sdir(char *dir)
{
	sdirs = resizemem(sdirs, (nsdirs + 1) * sizeof(char *));
	sdirs[nsdirs++] = stringcopy(dir);
}

void main(int argc, char **argv)
{
	int i;
	char *ifn = "-";
	char buf[FILENAME_MAX];

	myname = argv[0];

	add_def("XCOMM", "#");
	add_def("@@", "\n");
	add_def(SYSTEM, "1");
	add_def(ARCH, "1");
	add_def("__DEAVEN_LIB_HOME__", DMDDIR);
	add_def("__UTIL_TOP_DIR__", USR_TOPDIR);
#if DEVELOPMENT
	add_def("__DEVELOPMENT__", "1");
#endif

	add_sdir(".");
	sprintf(buf, "%s/include", DMDDIR);
	add_sdir(buf);
#ifdef BOTH_DMD_TOP_DEFINED  /* and could be different... */
	sprintf(buf, "%s/include", USR_TOPDIR);
	add_sdir(buf);
#endif /* BOTH_DMD_TOP_DEFINED */

	for(i=1; i<argc; ++i) {
	  if(argv[i][0] == '-') {
	    switch(argv[i][1]) {

	      case 'I':
		add_sdir(argv[i] + 2);
		break;

	      case 'D':
		{
		  char *s, *equ;
		  s = stringcopy(argv[i] + 2);
		  equ = strchr(s, '=');
		  if(equ) {
		    *equ = '\0';
		    add_def(s, equ + 1);
		  }
		  else
		    add_def(s, "1");
		  free(s);
		}
		break;

	      default:
		fprintf(stderr, usage, argv[0]);
		exit(1);
	    }
	  }
	  else {
	    ifn = argv[i];
	  }
	}

	do_file(ifn);
}
