/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: eishtridi.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <math.h>
#include "eisdiag.h"

/* Translated from Fortran 77 to C by f2c(1),
 * cleaned up by David Deaven 6 April 1994 
 */

#define PYTHAG(a,b) (sqrt(a*a+b*b))

void eishtridi(int n, double *ar, double *ai,
	double *d, double *e, double *e2, double *tau)
{
    /* System generated locals */
    int i__2, i__3;
    double d__1, d__2;

    /* Local variables */
    double f, g, h;
    int i, j, k, l;
    double scale, fi, gi, hh;
    int ii;
    double si;
    int jp1;
#pragma disjoint(*ar,*ai,*d,*tau)

/*     This subroutine is a translation of a complex analogue of 
 *     the algol procedure tred1, num. math. 11, 181-195(1968)
 *     by Martin, Reinsch, and Wilkinson. 
 *     handbook for auto. comp., vol.ii-linear algebra, 212-226(1971). 
 *
 *     This subroutine reduces a complex hermitian matrix 
 *     to a real symmetric tridiagonal matrix using 
 *     unitary similarity transformations. 
 *
 *     On input 
 *
 *        n is the order of the matrix. 
 *
 *        ar and ai contain the real and imaginary parts, 
 *          respectively, of the complex hermitian input matrix. 
 *          only the lower triangle of the matrix need be supplied. 
 *
 *     On output 
 *
 *        ar and ai contain information about the unitary trans- 
 *          formations used in the reduction in their full lower 
 *          triangles.  their strict upper triangles and the 
 *          diagonal of ar are unaltered. 
 *
 *        d contains the diagonal elements of the the tridiagonal matrix. 
 *
 *        e contains the subdiagonal elements of the tridiagonal 
 *          matrix in its last n-1 positions.  e(1) is set to zero. 
 *
 *        e2 contains the squares of the corresponding elements of e. 
 *          e2 may coincide with e if the squares are not needed. 
 *
 *        tau contains further information about the transformations. 
 *
 *     Questions and comments should be directed to Burton S. Garbow,
 *     mathematics and computer science div, argonne national laboratory 
 *
 *     This version dated august 1983.
 */

    /* Parameter adjustments */
    tau -= 3;
    --e2;
    --e;
    --d;
    ai -= (n + 1);
    ar -= (n + 1);

    /* Function Body */
    tau[(n << 1) + 1] = 1.;
    tau[(n << 1) + 2] = 0.;

    for (i = 1; i <= n; ++i) {
/* L100: */
	d[i] = ar[i + i * n];
    }
/*     .......... for i=n step -1 until 1 do -- .......... */
    for (ii = 1; ii <= n; ++ii) {
	i = n + 1 - ii;
	l = i - 1;
	h = 0.;
	scale = 0.;
	if (l < 1) {
	    goto L130;
	}
/*     .......... scale row (algol tol then not needed) .......... */
	i__2 = l;
	for (k = 1; k <= i__2; ++k) {
/* L120: */
	    scale = scale + (d__1 = ar[i + k * n], fabs(d__1)) + (d__2 = 
		    ai[i + k * n], fabs(d__2));
	}

	if (scale != 0.) {
	    goto L140;
	}
	tau[(l << 1) + 1] = 1.;
	tau[(l << 1) + 2] = 0.;
L130:
	e[i] = 0.;
	e2[i] = 0.;
	goto L290;

L140:
	i__2 = l;
	for (k = 1; k <= i__2; ++k) {
	    ar[i + k * n] /= scale;
	    ai[i + k * n] /= scale;
	    h += ar[i + k * n] * ar[i + k * n] +
	         ai[i + k * n] * ai[i + k * n];
/* L150: */
	}

	e2[i] = scale * scale * h;
	g = sqrt(h);
	e[i] = scale * g;
	f = PYTHAG(ar[i + l * n], ai[i + l * n]);
/*     .......... form next diagonal element of matrix t .......... */

	if (f == 0.) {
	    goto L160;
	}
	tau[(l << 1) + 1] = (
	  ai[i + l * n] * tau[(i << 1) + 2] -
	  ar[i + l * n] * tau[(i << 1) + 1]   ) / f;
	si = (
	  ar[i + l * n] * tau[(i << 1) + 2] +
	  ai[i + l * n] * tau[(i << 1) + 1]   ) / f;
	h += f * g;
	g = g / f + 1.;
	ar[i + l * n] = g * ar[i + l * n];
	ai[i + l * n] = g * ai[i + l * n];
	if (l == 1) {
	    goto L270;
	}
	goto L170;
L160:
	tau[(l << 1) + 1] = -tau[(i << 1) + 1];
	si = tau[(i << 1) + 2];
	ar[i + l * n] = g;
L170:
	f = 0.;

	i__2 = l;
	for (j = 1; j <= i__2; ++j) {
	    g = 0.;
	    gi = 0.;
/*     .......... form element of a*u .......... */
	    i__3 = j;
	    for (k = 1; k <= i__3; ++k) {
		g += ar[j + k * n] * ar[i + k * n] +
		     ai[j + k * n] * ai[i + k * n];
		gi += - ar[j + k * n] * ai[i + k * n]
		      + ai[j + k * n] * ar[i + k * n];
/* L180: */
	    }

	    jp1 = j + 1;
	    if (l < jp1) {
		goto L220;
	    }

	    i__3 = l;
	    for (k = jp1; k <= i__3; ++k) {
		g = g + ar[k + j * n] * ar[i + k * n] - ai[k + j *
			 n] * ai[i + k * n];
		gi = gi - ar[k + j * n] * ai[i + k * n] - ai[k + 
			j * n] * ar[i + k * n];
/* L200: */
	    }
/*     .......... form element of p .......... */
L220:
	    e[j] = g / h;
	    tau[(j << 1) + 2] = gi / h;
	    f +=  e[j] * ar[i + j * n] -
	          tau[(j << 1) + 2] * ai[i + j * n];
/* L240: */
	}

	hh = f / (h + h);
/*     .......... form reduced a .......... */
	i__2 = l;
	for (j = 1; j <= i__2; ++j) {
	    f = ar[i + j * n];
	    g = e[j] - hh * f;
	    e[j] = g;
	    fi = -ai[i + j * n];
	    gi = tau[(j << 1) + 2] - hh * fi;
	    tau[(j << 1) + 2] = -gi;

	    i__3 = j;
	    for (k = 1; k <= i__3; ++k) {
		ar[j + k * n] += 
		  - f * e[k] - g * ar[ i + k * n]
		  + fi * tau[(k << 1) + 2] + gi * ai[i + k * n];
		ai[j + k * n] += 
		  - f * tau[(k << 1) + 2] - g * ai[i + k * n]
		  - fi * e[k] - gi * ar[ i + k * n];
/* L260: */
	    }
	}

L270:
	i__3 = l;
	for (k = 1; k <= i__3; ++k) {
	    ar[i + k * n] = scale * ar[i + k * n];
	    ai[i + k * n] = scale * ai[i + k * n];
/* L280: */
	}

	tau[(l << 1) + 2] = -si;
L290:
	hh = d[i];
	d[i] = ar[i + i * n];
	ar[i + i * n] = hh;
	ai[i + i * n] = scale * sqrt(h);
/* L300: */
    }
}

