/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: eistql1.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <math.h>
#include "eisdiag.h"

/* Translated from Fortran 77 to C by f2c(1),
 * cleaned up by David Deaven 6 April 1994 
 */

#define PYTHAG(a,b) (sqrt(a*a+b*b))
#define d_sign(a,b) (copysign(a,b))

int eistql1(int n, double *d, double *e)
{
    /* System generated locals */
    int i__2;
    double d__1, d__2;
    int ierr;

    /* Local variables */
    double c, f, g, h;
    int i, j, l, m;
    double p, r, s, c2, c3=0.0, s2=0.0;
    int l1, l2;
    int ii;
    double dl1, el1;
    int mml;
    double tst1, tst2;

/*     This subroutine is a translation of the algol procedure tql1, 
 *     num. math. 11, 293-306(1968) by Bowdler, Martin, Reinsch, and 
 *     Wilkinson. 
 *     handbook for auto. comp., vol.ii-linear algebra, 227-240(1971). 
 *
 *     This subroutine finds the eigenvalues of a symmetric 
 *     tridiagonal matrix by the ql method. 
 *
 *     On input 
 *
 *        n is the order of the matrix. 
 *
 *        d contains the diagonal elements of the input matrix. 
 *
 *        e contains the subdiagonal elements of the input matrix 
 *          in its last n-1 positions.  e(1) is arbitrary. 
 *
 *      On output 
 *
 *        d contains the eigenvalues in ascending order.  if an 
 *          error exit is made, the eigenvalues are correct and 
 *          ordered for indices 1,2,...ierr-1, but may not be 
 *          the smallest eigenvalues. 
 *
 *        e has been destroyed. 
 *
 *        return value is
 *          zero       for normal return, 
 *          j          if the j-th eigenvalue has not been 
 *                     determined after 30 iterations. 
 *
 *     Questions and comments should be directed to Burton S. Garbow, 
 *     mathematics and computer science div, argonne national laboratory 
 *
 *
 *     This version dated august 1983. 
 */

    /* Parameter adjustments */
    --e;
    --d;

    /* Function Body */
    ierr = 0;
    if (n == 1) {
	goto L1001;
    }

    for (i = 2; i <= n; ++i) {
/* L100: */
	e[i - 1] = e[i];
    }

    f = 0.;
    tst1 = 0.;
    e[n] = 0.;

    for (l = 1; l <= n; ++l) {
	j = 0;
	h = (d__1 = d[l], fabs(d__1)) + (d__2 = e[l], fabs(d__2));
	if (tst1 < h) {
	    tst1 = h;
	}
/*     .......... look for small sub-diagonal element .......... */
	i__2 = n;
	for (m = l; m <= i__2; ++m) {
	    tst2 = tst1 + (d__1 = e[m], fabs(d__1));
	    if (tst2 == tst1) {
		goto L120;
	    }
/*     .......... e(n) is always zero, so there is no exit */
/*                through the bottom of the loop .......... */
/* L110: */
	}

L120:
	if (m == l) {
	    goto L210;
	}
L130:
	if (j == 30) {
	    goto L1000;
	}
	++j;
/*     .......... form shift .......... */
	l1 = l + 1;
	l2 = l1 + 1;
	g = d[l];
	p = (d[l1] - g) / (e[l] * 2.);
	r = PYTHAG(p, 1.0);
	d[l] = e[l] / (p + d_sign(r, p));
	d[l1] = e[l] * (p + d_sign(r, p));
	dl1 = d[l1];
	h = g - d[l];
	if (l2 > n) {
	    goto L145;
	}

	i__2 = n;
	for (i = l2; i <= i__2; ++i) {
/* L140: */
	    d[i] -= h;
	}

L145:
	f += h;
/*     .......... ql transformation .......... */
	p = d[m];
	c = 1.;
	c2 = c;
	el1 = e[l1];
	s = 0.;
	mml = m - l;
/*     .......... for i=m-1 step -1 until l do -- .......... */
	i__2 = mml;
	for (ii = 1; ii <= i__2; ++ii) {
	    c3 = c2;
	    c2 = c;
	    s2 = s;
	    i = m - ii;
	    g = c * e[i];
	    h = c * p;
	    r = PYTHAG(p, e[i]);
	    e[i + 1] = s * r;
	    s = e[i] / r;
	    c = p / r;
	    p = c * d[i] - s * g;
	    d[i + 1] = h + s * (c * g + s * d[i]);
/* L200: */
	}

	p = -s * s2 * c3 * el1 * e[l] / dl1;
	e[l] = s * p;
	d[l] = c * p;
	tst2 = tst1 + (d__1 = e[l], fabs(d__1));
	if (tst2 > tst1) {
	    goto L130;
	}
L210:
	p = d[l] + f;
/*     .......... order eigenvalues .......... */
	if (l == 1) {
	    goto L250;
	}
/*     .......... for i=l step -1 until 2 do -- .......... */
	i__2 = l;
	for (ii = 2; ii <= i__2; ++ii) {
	    i = l + 2 - ii;
	    if (p >= d[i - 1]) {
		goto L270;
	    }
	    d[i] = d[i - 1];
/* L230: */
	}

L250:
	i = 1;
L270:
	d[i] = p;
/* L290: */
    }

    goto L1001;
/*     .......... set error -- no convergence to an */
/*                eigenvalue after 30 iterations .......... */
L1000:
    ierr = l;
L1001:
    return ierr;
}

