/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: eistred2.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

/*     this subroutine is a translation of the algol procedure tred2,
 *     num. math. 11, 181-195(1968) by martin, reinsch, and wilkinson.
 *     handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).
 *
 *     this subroutine reduces a real symmetric matrix to a
 *     symmetric tridiagonal matrix using and accumulating
 *     orthogonal similarity transformations.
 *
 *     on input
 *
 *        nm must be set to the row dimension of two-dimensional
 *          array parameters as declared in the calling program
 *          dimension statement.
 *
 *        n is the order of the matrix.
 *
 *        a contains the real symmetric input matrix.  only the
 *          lower triangle of the matrix need be supplied.
 *
 *     on output
 *
 *        d contains the diagonal elements of the tridiagonal matrix.
 *
 *        e contains the subdiagonal elements of the tridiagonal
 *          matrix in its last n-1 positions.  e(1) is set to zero.
 *
 *        z contains the orthogonal transformation matrix
 *          produced in the reduction.
 *
 *        a and z may coincide.  if distinct, a is unaltered.
 *
 *     questions and comments should be directed to burton s. garbow,
 *     mathematics and computer science div, argonne national laboratory 
 *
 *     this version dated august 1983.
 */

#include <math.h>
#include "eisdiag.h"

int eistred2(double *z, int n, double *d, double *e)
{
    /* Local variables */
    double f, g, h;
    int i, j, k, l;
    double scale, hh;
    int jp1;
#pragma disjoint(*z,*d,*e)


    /* Function Body */
    for (i = 0; i < n; ++i) 
	d[i] = z[n-1 + i * n];

    if (n > 1) 
    {
/*     .......... for i=n step -1 until 2 do -- .......... */
    for (i = n-1; i > 0; i--) 
    {
	l = i - 1;
	h = 0.;
	scale = 0.;
	if (l >0) 
	{
	    for (k = 0; k <= l; ++k) 
		scale += fabs(d[k]);

	    if (scale != 0.) 
		goto L140;
	}

	e[i] = d[l];

	for (j = 0; j <= l; ++j) 
	{
	    d[j] = z[l + j * n];
	    z[i + j * n] = 0.;
	    z[j + i * n] = 0.;
	}

	goto L290;

L140:
	for (k = 0; k <= l; ++k) 
	{
	    d[k] /= scale;
	    h += d[k] * d[k];
	}

	f = d[l];
	g = -copysign(sqrt(h),f);
	e[i] = scale * g;
	h -= f * g;
	d[l] = f - g;
/*     .......... form a*u .......... */
	for (j = 0; j <= l; ++j) 
	    e[j] = 0.;

	for (j = 0; j <= l; ++j) 
	{
	    f = d[j];
	    z[j + i * n] = f;
	    g = e[j] + z[j + j * n] * f;
	    jp1 = j + 1;
	    if (l >= jp1) 
	    {
		for (k = jp1; k <= l; ++k) 
		{
		    g += z[j*n+k] * d[k];
		    e[k] += z[j*n+k] * f;
		}
	    }
	    e[j] = g;
	}
/*     .......... form p .......... */
	f = 0.;

	for (j = 0; j <= l; ++j) 
	{
	    e[j] /= h;
	    f += e[j] * d[j];
	}

	hh = f / (h + h);
/*     .......... form q .......... */
	for (j = 0; j <= l; ++j) 
	    e[j] -= hh * d[j];

/*     .......... form reduced a .......... */
	for (j = 0; j <= l; ++j) 
	{
	    f = d[j];
	    g = e[j];
	    for (k = j; k <= l; ++k) 
		z[j*n+k] -=  f * e[k] + g * d[k];

	    d[j] = z[j*n+l];
	    z[j*n+i] = 0.;
	}

L290:
	d[i] = h;

    }
    
/*     .......... accumulation of transformation matrices .......... */
    for (i = 1; i < n; ++i) 
    {
	l = i - 1;
	z[n-1 + l * n] = z[l + l * n];
	z[l + l * n] = 1.;
	h = d[i];
	if (h == 0.)
	    goto L380;

	for (k = 0; k <= l; ++k) 
	    d[k] = z[k + i * n] / h;

	for (j = 0; j <= l; ++j) 
	{
	    g = .0;
	    for (k = 0; k <= l; k++)
		g += z[k+i*n] * z[k+j*n];

	    for (k = 0; k <= l; ++k) 
		z[k+j*n] -= g * d[k];
	}

L380:
	for (k = 0; k <= l; ++k) 
	    z[k + i * n] = 0.;
    }
}

    for (i = 0; i < n; ++i) 
    {
	d[i] = z[i * n+n-1];
	z[i * n+n-1] = 0.;
    }

    z[n*n-1] = 1.;
    e[0] = 0.;
    return 0;
} /* tred2_ */
