/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: fit.c,v 1.2 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "memory.h"
#include "fit.h"

static double expt_err(Fit fit, int which)
{
	double unit = fabs(fit->exptdata[which]);
	if(unit < 1.e-10) unit = 1.0;
	return (fit->expt[which] - fit->exptdata[which]) / unit;
}

/* Compute the RMS error from the current experimental values.
 */
static void compute_rms(Fit fit)
{
	int i;
	double err, weight;

	fit->rms = weight = 0.0;

	for(i=0; i<fit->nexpt; ++i)
	{
	  err = expt_err(fit, i);
	  fit->rms += fit->exptweight[i] * err * err;
	  weight += fit->exptweight[i];
	}
	fit->rms = sqrt(fit->rms / weight);
}

Fit NewFit(int npar, int nexpt)
{
	int i;
	char buf[64];
	Fit fit = NewPointer(fit);

	fit->npar = npar;
	fit->fitpar = imalloc(npar);
	fit->parname = NewPointerArray(fit->parname, npar);
	fit->par = dmalloc(npar);
	fit->parmin = dmalloc(npar);
	fit->parmax = dmalloc(npar);
	fit->changed = 1;
	for(i=0; i<npar; i++) {
	  fit->fitpar[i] = 1;
	  sprintf(buf, "p%d", i);
	  fit->parname[i] = mystrdup(buf);
	  fit->par[i] = 0.0;
	  fit->parmin[i] = NO_RESTRICTION;
	  fit->parmax[i] = NO_RESTRICTION;
	}

	fit->nexpt = nexpt;
	fit->exptweight = dmalloc(nexpt);
	fit->exptname = NewPointerArray(fit->exptname, nexpt);
	fit->exptdata = dmalloc(nexpt);
	fit->expt = dmalloc(nexpt);
	for(i=0; i<nexpt; i++) {
	  fit->exptweight[i] = 1.0;
	  sprintf(buf, "e%d", i);
	  fit->exptname[i] = mystrdup(buf);
	  fit->exptdata[i] = 1.0;
	  fit->expt[i] = 0.0;
	}

	fit->rms = 1.0;
	fit->model = NULL;
	fit->resources = NULL;

	return fit;
}

void DestroyFit(Fit fit)
{
	int i;
	if(fit) {
	  for(i=0; i<fit->npar; i++)
	    myfree(fit->parname[i]);
	  myfree(fit->fitpar);
	  myfree(fit->parname);
	  myfree(fit->par);
	  myfree(fit->parmin);
	  myfree(fit->parmax);

	  for(i=0; i<fit->nexpt; i++)
	    myfree(fit->exptname[i]);
	  myfree(fit->exptweight);
	  myfree(fit->exptname);
	  myfree(fit->exptdata);
	  myfree(fit->expt);

	  myfree(fit);
	}
}


int FitEvaluate(Fit fit)
{
	int r = 0;
	if(fit->changed) {
	  r = fit->model(fit->par, fit->expt, fit->resources);
	  fit->changed = 0;
	}
	compute_rms(fit);
	return r;
}

/* Print the current state of the fit parameters to a file.
 */
void FitPrint(FILE *f, Fit fit)
{
	int i;

	FitEvaluate(fit);

	for(i=0; i<fit->npar; ++i)
	  fprintf(f, "%+14.8e%s\t%s\n", fit->par[i],
	    fit->fitpar[i] ? "<" : " ", fit->parname[i]);

	fprintf(f, "--Experiment--  -----Model----  ---Error---  -Weight-\n");
	for(i=0; i<fit->nexpt; ++i)
	{
	  fprintf(f, "%+14.8g  %+14.8g  %+10.2f%%  %8f %s\n", fit->exptdata[i],
	    fit->expt[i], 100.0 * expt_err(fit, i), fit->exptweight[i],
	    fit->exptname[i]);
	}
	fprintf(f, "RMS error = %.5g\n", fit->rms);
}

/* Return a trial range for fit parameter which.
 */
double FitTrialRange(Fit fit, int which)
{
	double pmin, pmax, pabs = fabs(fit->par[which]);

	pmin = fit->parmin[which] == NO_RESTRICTION ?
	  fit->par[which] - pabs : fit->parmin[which];
	pmax = fit->parmax[which] == NO_RESTRICTION ?
	  fit->par[which] + pabs : fit->parmax[which];

	return (pmin == pmax ? 1.0 : pmax - pmin);
}
