/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: fitmain.c,v 1.1 1996/05/22 19:38:24 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "dmd/memory.h"
#include "dmd/tools.h"
#include "dmd/fit.h"

#define NO_RESTRICTION_STRING "none"

extern int verbose;

static double scan_limit(FILE *f, int *lastc)
{
	int i, c;
	char buf[128];

	for(i=0; i<128 && (c = fgetc(f)) != EOF; i++) {
	  *lastc = c;
	  if(strchr("(),:- \t\n", c)) {
	    buf[i] = '\0';
	    if(!strcmp(buf, NO_RESTRICTION_STRING)) return NO_RESTRICTION;
	    return atof(buf);
	  }
	  buf[i] = c;
	}
	buf[i] = '\0';
	return atof(buf); /* should never happen */
}

static void scan_to(FILE *f, int until)
{
	int c;
	while((c = fgetc(f)) != until && c != EOF);
}

/* Read parameter and experiment values from the file.
 */
static void scan_fit(FILE *f, Fit fit)
{
	int i, c;
	char pc;

	for(i=0; i<fit->npar; i++) {
	  fscanf(f, "%lf(", fit->par + i);
	  fit->parmin[i] = scan_limit(f, &c);
	  if(c != ',') scan_to(f, ',');
	  fit->parmax[i] = scan_limit(f, &c);
	  if(c != ')') scan_to(f, ')');
	  pc = ' ';
	  fscanf(f, "%c", &pc);
	  fit->fitpar[i] = (isspace(pc) ? 0 : 1);
	  scan_to(f, '\n');
	}

	for(i=0; i<fit->nexpt; i++) {
	  fscanf(f, "%lf %lf", fit->exptdata + i, fit->exptweight + i);
	  scan_to(f, '\n');
	}

	fit->changed = 1;
}

static void print_limit(FILE *f, double limit)
{
	if(limit == NO_RESTRICTION)
	  fprintf(f, NO_RESTRICTION_STRING);
	else
	  fprintf(f, "%g", limit);
}

/* Print the current state of the fit parameters to a file.
 */
static void print_fit(FILE *f, Fit fit)
{
	int i;

	for(i=0; i<fit->npar; i++) {
	  fprintf(f, "%18.12g(", fit->par[i]);
	  print_limit(f, fit->parmin[i]);
	  fprintf(f, ",");
	  print_limit(f, fit->parmax[i]);
	  fprintf(f, ")%c %s\n",
	    fit->fitpar[i] ? '#' : ' ', fit->parname[i]);
	}

	fprintf(f, "\n");

	for(i=0; i<fit->nexpt; i++)
	  fprintf(f, "%12g %5f %s\n",
	    fit->exptdata[i], fit->exptweight[i], fit->exptname[i]);
}

int FitMain(Fit fit, int argc, char **argv)
{
	int i;
	int (*method)(Fit, char *, FILE *) = FitByConjugateGradient;
	char *options = NULL;

	static struct {
	  char *name;
	  int (*method)(Fit, char *, FILE *);
	}
	method_table[] = {
	  {"ConjugateGradient",		FitByConjugateGradient},
	  {"cg",			FitByConjugateGradient},
	  {"MonteCarlo",		FitByMonteCarlo},
	  {"mc",			FitByMonteCarlo},
	  {"DownhillSimplex",		FitByDownhillSimplex},
	  {"ds",			FitByDownhillSimplex},
	  {"GridSearch",		FitByGridSearch},
	  {"gs",			FitByGridSearch},
	  {NULL, NULL}
	};
	static char *usage = "usage: %s [-v] [-m method] <options>\n";

	for(i=1; i<argc; i++) {
	  if(argv[i][0] == '-') {
	    switch(argv[i][1]) {
	      case 'm':
		i++;
		{ int which;
		  for(which=0; method_table[which].name; which++) {
		    if(!strcmp(argv[i], method_table[which].name))
		    {
		      method = method_table[which].method;
		      printf("Using %s method\n", method_table[which].name);
		    }
		  }
		}
		break;

	      case 'v':
		verbose = 1;
		break;

	      case 'V':
	      case 'h':
		fprintf(stderr, usage, argv[0]);
		return 0;

	      default:
		options = ResizePointerArray(options, 
		  (options ? strlen(options) : 0) + strlen(argv[i]) + 2);
		strcat(options, argv[i]);
		strcat(options, " ");
		break;
	    }
	  }
	}

	{ FILE *f;
	  char *fn = "fit.in";
	  if((f = fopen(fn, "r"))) {
	    scan_fit(f, fit);
	    fclose(f);
	    if(verbose) printf("%s: read fit from `%s'\n", argv[0], fn);
	  }
	  else {
	    if(verbose)
	      printf("%s: could not open `%s', using default fit\n",
		argv[0], fn);
	  }
	}

	FitPrint(stdout, fit);
	method(fit, options ? options : "", stdout);
	FitPrint(stdout, fit);

	{ char *fn = "fit.out";
	  FILE *f = fopen_with_default(fn, "w", NULL);
	  print_fit(f, fit);
	  fclose(f);
	  if(verbose) printf("%s: wrote fit to `%s'\n", argv[0], fn);
	}

	return 0;
}
