/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: galignypr.c,v 1.3 1996/07/09 20:11:00 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "dmd/geometry.h"
#include "dmd/memory.h"
#include "dmd/tools.h"

/* Find the best re-ordering of atoms in g2 to align with g1.
 * c1 and c2 are the box coordinate centers of g1 and g2.
 * g1x0 is the physical coordinates of g1, and g2x0 is scratch space
 * for the coordinates of g2.
 * t is the physical rotation matrix to apply to g2.
 * used is scratch space to record the re-numbering of g2's atoms.
 */
static double gdiff_x(Geometry *g1, double c1[3], Geometry *g2, double c2[3],
	double *g1x0, double *g2x0, double t[9], int *used)
{
	int i, j, m, jmin = 0;
	double rms = 0.0, d, dmin, a, c1p[3], c2p[3], tbox[9];

	for(i=0; i<(g2->n); ++i) used[i] = -1;

	mmul3(g1->box, c1, c1p);

	/* Make a single transformation tbox = (t box) to convert box
	 * coordinates to physical coordinates and rotate them.
	 * Apply the transformations to all atoms in g2.
	 */
	matmul3(t, g2->box, tbox);
	mmul3(tbox, c2, c2p); /* rotated center */
	vmmul3(tbox, g2->x, g2x0, g2->n); /* rotated physical coordinates */

	/* For each atom in g1, find nearest unused atom in g2 and use it
	 * to compute the contribution to the RMS distance.
	 */
	for(i=0; i<(g1->n); ++i) {
	  dmin = -1.0;
	  jmin = -1;
	  for(j=0; j<(g2->n); ++j) {
	    if(used[j] < 0) {
	      d = 0.0;
	      for(m=0; m<3; ++m) {
		a = (g1x0[3*i + m] - c1p[m]) - (g2x0[3*j + m] - c2p[m]);
		d += a * a;
	      }
	      if(dmin < 0.0 || d < dmin) {
		dmin = d;
		jmin = j;
	      }
	    }
	  }
	  if(jmin >= 0) {
	    used[jmin] = i;
	    rms += dmin;
	  }
	}

	rms = sqrt(rms / g1->n);
	return rms;
}

/* Get the physical radius and the box center.
 */
static double center_and_radius(Geometry *g, double c[3])
{
	int i, m;
	double r2, x[3], r = 0.0;

	c[0] = c[1] = c[2] = 0.0;

	for(i=0; i<(3 * g->n); ++i) c[i % 3] += g->x[i];

	for(m=0; m<3; ++m) 
	  c[m] /= g->n;

	for(i=0; i<(g->n); ++i)
	{
	  for(m=0; m<3; ++m) x[m] = g->x[i*3 + m] - c[m];
	  mtrn3(g->box, x);
	  r2 = x[0] * x[0] + x[1] * x[1] + x[2] * x[2];
	  if(r2 > r) r = r2;
	}

	r = sqrt(r);
	return r;
}

/* Given a geometry g1, determine the best YPR rotation in physical
 * space (ypr[]) followed by a physical coordinate translation (dx[])
 * to superimpose g2 on g1.  Return the RMS difference in physical
 * atomic positions.
 * Use a grid search with ngrid points and a spacial resolution of
 * res in the final answer.
 */
double GAlignYPR(Geometry *g1, Geometry *g2,
	double ypr[3], double dx[3], int ngrid, double res)
{
	int m;
	int *used, iypr[3], iyprmin[3], done;
	double *g1x0, *g2x0, c1[3], c2[3];
	double rms, rmsmin;
	double dypr[3], ypr0[3], ypr1[3], t[3*3];
	double r1, r2;

	/* We need some scratch space */
	used = imalloc(g2->n);
	g1x0 = dmalloc(3 * g2->n);
	g2x0 = dmalloc(3 * g2->n);

	/* Find the center of number and the radius for the geometries.
	 * The translate amount is the difference in centers.
	 */
	r1 = center_and_radius(g1, c1);
	r2 = center_and_radius(g2, c2);
	res = res / (r1 > r2 ? r1 : r2); /* convert to angular measure */

	/* Keep a copy of g1's physical coordinates.
	 */
	vmmul3(g1->box, g1->x, g1x0, g1->n);

	/* Start with the entire range, then narrow the search
	 * until we get the desired resolution.
	 */
	for(m=0; m<3; ++m) {
	  ypr0[m] = 0.0;
	  iyprmin[m] = 0;
	  dypr[m] = M_PI;
	}

	do {
	  done = 1;
	  for(m=0; m<3; ++m) {
	    ypr0[m] = ypr0[m] + (iyprmin[m] - 1) * dypr[m];
	    ypr1[m] = ypr0[m] + 2.0 * dypr[m];
	    dypr[m] = (ypr1[m] - ypr0[m]) / ngrid;
	    if(dypr[m] > res) done = 0;
	  }

	  rmsmin = -1.0;

	  /* Explore all rotations, determine minimum rotation
	   * direction.  This is a triple search loop (!).
	   */
	  for(iypr[0]=0; iypr[0]<=ngrid; ++iypr[0]) {
	    ypr[0] = ypr0[0] + iypr[0] * dypr[0];
	    for(iypr[1]=0; iypr[1]<=ngrid; ++iypr[1]) {
	      ypr[1] = ypr0[1] + iypr[1] * dypr[1];
	      for(iypr[2]=0; iypr[2]<=ngrid; ++iypr[2]) {
		ypr[2] = ypr0[2] + iypr[2] * dypr[2];

		YPRRotationMatrix(ypr[0], ypr[1], ypr[2], t);
		rms = gdiff_x(g1, c1, g2, c2, g1x0, g2x0, t, used);

		if(rmsmin < 0.0 || rms < rmsmin)
		{
		  rmsmin = rms;
		  for(m=0; m<3; ++m) iyprmin[m] = iypr[m];
		}
	      }
	    }
	  }

	} while(!done);

	/* Store the best answer.
	 */
	for(m=0; m<3; ++m) ypr[m] = ypr0[m] + iyprmin[m] * dypr[m];
	YPRRotationMatrix(ypr[0], ypr[1], ypr[2], t);
	mtrn3(g2->box, c2);
	mtrn3(t, c2);
	mtrn3(g1->box, c1);
	for(m=0; m<3; ++m) dx[m] = c1[m] - c2[m];

	myfree(used);
	myfree(g1x0);
	myfree(g2x0);

	return rmsmin;
}
