/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: galloc.c,v 1.4 1995/10/03 21:00:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "geometry.h"
#include "memory.h"

/* Make a geometry with duplicate pointers.
 */
Geometry *Gdup(Geometry *src)
{
  Geometry *dst = mymalloc(sizeof(Geometry));
  dst->n = src->n;
  dst->x = src->x;
  dst->box = src->box;
  dst->boxi = src->boxi;
  dst->name = src->name;
  dst->aname = src->aname;
  dst->period = src->period;
  dst->flags = src->flags;
  return dst;
}

/* Allocate memory for a geometry description
 * Assume no flags.
 */
Geometry *Gmalloc(int n, char *name)
{
  int i, j;
  Geometry *new;
  new = mymalloc(sizeof(Geometry));
  new->n = n;
  new->x = n > 0 ? dmalloc(3*n) : NULL;
  new->box = dmalloc(9);
  new->boxi = dmalloc(9);
  new->name = mystrdup(name);
  new->aname = n > 0 ? mymalloc(n * sizeof(char *)) : NULL;
  new->flags = NULL;
  for(i=0; i<n; ++i) new->aname[i] = NULL;
  new->period = 0;
  /* Optimizers have trouble with shorter versions of this because
   * (I think) box and boxi are sequential in a structure. XXX
   */
  for(i=0; i<3; ++i) 
    for(j=0; j<3; ++j) 
      new->box[j+3*i] = (i==j ? 1.0 : 0.0);
  for(i=0; i<9; ++i) new->boxi[i] = new->box[i];
  return new;
}

/* Re-allocate memory for a geometry description, preserving
 * the contents.  ALWAYS returns old, unless old is NULL.
 * If the geometry gets bigger, the extra atoms have indeterminate
 * information stored for the coordinates.
 */
Geometry *Grealloc(Geometry *old, int n)
{
  int i;
  if(old == NULL) {
    old = Gmalloc(n, NULL);
  }
  else if(n != old->n) {
    /* We have to handle the atomic names carefully, freeing up
     * each string before resizeing the `aname' array, and
     * initializing new strings to NULL after resizing.
     */
    for(i=n; i<(old->n); ++i) myfree(old->aname[i]);
    if(n > 0) {
      old->x = myrealloc(old->x, 3 * n * sizeof(double));
      old->aname = myrealloc(old->aname, n * sizeof(char *));
    }
    else {
      myfree(old->x);
      old->x = NULL;
      myfree(old->aname);
      old->aname = NULL;
    }
    for(i=(old->n); i<n; ++i) old->aname[i] = NULL;

    if(old->flags != NULL)
    {
      if(n > 0) {
	old->flags = myrealloc(old->flags, n * sizeof(int));
	for(i=(old->n); i<n; ++i) old->flags[i] = 0;
      }
      else {
	myfree(old->flags);
	old->flags = NULL;
      }
    }
    old->n = n;
  }
  return old;
}

void Gfree(Geometry *g)
{
	int i;
	if(g != NULL) {
	  myfree(g->x);
	  myfree(g->box);
	  myfree(g->boxi);
	  myfree(g->name);
	  for(i=0; i<(g->n); ++i) myfree(g->aname[i]);
	  myfree(g->aname);
	  myfree(g->flags);
	  myfree(g);
	}
}

