/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gentra.c,v 1.5 1996/01/29 21:25:58 deaven Exp $
 * $Author: deaven $
 */

#include "tools.h"

/*
 * Generate a transformation matrix for a virtual camera
 *
 * Input:
 *
 * xcam(3)   -- position of camera.
 * vcam(3)   -- view direction of camera.
 * up(3)     -- direction facing upwards (if same as vcam, an
 *              unknown camera orientation is chosen).
 *
 * Output:
 *
 * t(4,4)    -- transformation matrix.  multiply a coordinate
 *              column vector by t to get the coordinates in the
 *              camera coordinate system, which has x to the right,
 *              y upwards, and z towards the viewer (assuming a right
 *              handed coordinate system).
 * scam(9)   -- camera coordinate system vectors in the original
 *              coordinate system.
 *		  I.e. scam[0,1,2] = x direction for camera
 */
void Gentra(double xcam[3], double vcam[3], double up[3],
	double t[4*4], double scam[3*3])
{
  int i, j;
  double zz, s;

/* z is -vcam. if zero generate default vcam (1 0 0). */

  zz=0.0;
  for(i=0; i<3; ++i) {
    zz += vcam[i]*vcam[i];
    scam[6+i] = -vcam[i];  /* last original vector */
  }
  if(zz <= 0.0) {
    scam[6] = -1.0; scam[5] = scam[4] = 0.0;
  }

/* x is up cross z (if zero then pick one). */

  Cross(up, scam+6, scam+0);
  zz = scam[0]*scam[0] + scam[1]*scam[1] + scam[2]*scam[2];
  if(zz <= 0.0) {
    scam[3] = scam[6] + 1.0;
    scam[4] = scam[7];
    scam[5] = scam[8];
    Cross(scam+3, scam+6, scam+0);
  }

/* y is z cross x. */

  Cross(scam+6, scam+0, scam+3);

/* normalize the components */

  for(i=0; i<3; ++i) {
    s = 0.0;
    for(j=0; j<3; ++j)s += scam[j+i*3]*scam[j+i*3];
    s = 1.0/sqrt(s);
    for(j=0; j<3; ++j) {
      scam[j+i*3] *= s;
      t[i+4*j] = scam[j+i*3];
    }
  }

/* for translation part, need to apply 3x3 rotation first to get it right */

  for(i=0; i<3; ++i) {
    t[i+4*3] = t[3+4*i] = 0.0;
    for(j=0; j<3; ++j)t[i+4*3] -= xcam[j]*t[i+4*j];
    t[3+4*3] = 1.0;
  }
}

