/* This file is part of David Deaven's personal library
 * Copyright (c) 1995 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gfile.c,v 1.4 1996/03/15 20:21:28 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "memory.h"
#include "geometry.h"

#define MARK_CHUNK 128		/* re-allocation value for marker table */

/* Private structure definition for the information about a
 * geometry stream.  "current" is where we are poised to read
 * the next geometry -- current==0 means the first one.
 */
struct Gfile {
	char *path;		/* path name of geometry file */
	FILE *f;		/* stream */
	fpos_t *file_marks;	/* file markers at each geometry */
	int known;		/* how many known markers */
	int current;		/* the current file position */
	int alloc;		/* allocated marks in table */
	int seekable;		/* is the file seekable? */
	time_t time;		/* last time file changed */
};

/* Add a mark to the end of the table, and record that the 
 * current position is there.
 */
static int add_mark(struct Gfile *gf)
{
	fpos_t cur;

	if(fgetpos(gf->f, &cur) < 0) return -1;

	if(gf->known + 1 > gf->alloc) {
	  gf->alloc += MARK_CHUNK;
	  gf->file_marks = myrealloc(gf->file_marks,
	    gf->alloc * sizeof(fpos_t));
	}

	gf->current = gf->known;
	gf->file_marks[gf->known++] = cur;
	return 0;
}

static struct Gfile *new(char *path)
{
	struct Gfile *gf = mymalloc(sizeof(*gf));

	gf->path = mystrdup(path);
	gf->f = NULL;
	gf->file_marks = NULL;
	gf->known = gf->current = gf->alloc = 0;
	gf->seekable = 0;
	gf->time = 0;

	return gf;
}

int GFileReopen(struct Gfile *gf)
{
	struct stat statbuf;

	if(gf->f) fclose(gf->f);
	gf->f = (strcmp(gf->path, "-") ? fopen(gf->path, "r") : stdin);
	if(gf->f == NULL) return -1;

	if(fstat(fileno(gf->f), &statbuf) < 0) return -1;
	gf->time = statbuf.st_mtime;

	myfree(gf->file_marks);
	  gf->file_marks = NULL;
	gf->known = gf->current = gf->alloc = 0;

	if(add_mark(gf) < 0)
	  gf->seekable = 0;
	else
	  gf->seekable = 1;

	return 0;
}

struct Gfile *GFileOpen(char *path)
{
	struct Gfile *gf;

	gf = new(path);
	if(GFileReopen(gf)) return NULL;

	return gf;
}

void GFileClose(struct Gfile *gf)
{
	if(gf) {
	  if(gf->f != NULL && gf->f != stdin) fclose(gf->f);
	  myfree(gf->file_marks);
	  myfree(gf->path);
	  myfree(gf);
	}
}

/* Read the next geometry from the attached file, incrementing the
 * current frame counter by one.
 */
Geometry *GFileRead(struct Gfile *gf)
{
	Geometry *g;
	struct stat statbuf;

restart:

	if(gf == NULL || gf->f == NULL || feof(gf->f)) return NULL;

	if((g = Gread(gf->f)) == NULL) {
	  /* Check to see if the file has been modified.  Reopen
	   * it if so.
	   */
	  if(fstat(fileno(gf->f), &statbuf) == 0) {
	    if(gf->time < statbuf.st_mtime) {
	      GFileReopen(gf);
	      goto restart;
	    }
	  }
	  return NULL;
	}

	++(gf->current);

	if(gf->current >= gf->known) add_mark(gf);

	return g;
}

/* Skip backwards by the given number of geometries.  Return
 * the actual number rewound.
 */
int GFileRewind(int n, struct Gfile *gf)
{
	return -GFileAdvance(-n, gf);
}

/* Skip forward n geometries. Return the actual number skipped,
 * which could be less than n if the end of the file was reached,
 * for example.  n may be negative.
 */
int GFileAdvance(int n, struct Gfile *gf)
{
	int target, actual = 0;
	Geometry *g;

	if(gf == NULL || n == 0) return actual;

	target = gf->current + n; /* what we want current to be */

	if(target < 0) target = 0;

	if(gf->seekable)
	{
	  if(target < gf->known) { /* we can go there directly */
	    fsetpos(gf->f, gf->file_marks + target);
	    actual += target - gf->current;
	    gf->current = target;
	  } else {
	    if(gf->known > 0 && gf->current < gf->known - 1) {
	      /* first go to the end of the known region */
	      fsetpos(gf->f, gf->file_marks + gf->known - 1);
	      actual += gf->known - 1 - gf->current;
	      gf->current = gf->known - 1;
	    }
	    while(gf->current < target) {
	      if((g = GFileRead(gf)) == NULL) break;
	      Gfree(g);
	      ++actual;
	    }
	  }
	}
	else /* seekable */
	{
	  if(n > 0) {
	    for(actual=0; actual<n; ++actual)
	      if((g = GFileRead(gf)) == NULL) break;
	  }
	}

	return actual;
}

void GFileInfo(struct Gfile *gf, FILE **f, char **path,
	int *frame, int *maxframe)
{
	if(f != NULL) *f = gf ? gf->f : NULL;
	if(path != NULL) *path = gf ? gf->path : NULL;
	if(frame != NULL) *frame = gf ? gf->current : 0;
	if(maxframe != NULL) *maxframe = gf ? gf->known : 0;
}
