/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gread.c,v 1.8 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "geometry.h"
#include "memory.h"

#define MODULE "Gread"
#define MAXLEN 256 /* maximum input line length */

#define ANGSTROM 0.529177  /* conversion from internal (atomic) */
#define ABS(x) ((x>0 ? x : -x))

/* Resizable buffer.
 */
struct Rbuf {
	unsigned long size;
	char *buf;
};

static int mystrlen(char *s)
{
	int len = 0;
	for(;(*s) != '\0'; ++s) ++len;
	return len;
}

static struct Rbuf *new_rbuf(unsigned long size)
{
	struct Rbuf *rb = mymalloc(sizeof(*rb));
	rb->buf = mymalloc(size * sizeof(*(rb->buf)));
	rb->size = size;
	if(size > 0) rb->buf[0] = '\0';
	return rb;
}

static struct Rbuf *resize_rbuf(struct Rbuf *rb, unsigned long newsize)
{
	if(rb == NULL)
	  rb = new_rbuf(newsize);
	else {
	  rb->buf = myrealloc(rb->buf, newsize * sizeof(*(rb->buf)));
	  rb->size = newsize;
	}
	return rb;
}

static void delete_rbuf(struct Rbuf *rb)
{
	if(rb != NULL) {
	  myfree(rb->buf);
	  myfree(rb);
	}
}

/* Read an arbitrary length string from a file.
 */
static struct Rbuf *fgets_rbuf(struct Rbuf *rb, FILE *f)
{
	unsigned long current = 0;
	char *ret;

	if(f == NULL || feof(f) || ferror(f)) return NULL;

	do {
	  if(rb == NULL || rb->size - current < MAXLEN)
	    rb = resize_rbuf(rb, (rb ? rb->size : 0) + MAXLEN);
	  ret = fgets(rb->buf + current, rb->size - current, f);
	  if(ret)
	    current = mystrlen(rb->buf);
	  else
	    rb->buf[current] = '\0';
	} while(ret && current > 1 && rb->buf[current - 1] != '\n');

	return rb;
}

static int gread_x(char *s, char **aname, double x[3], int *flag, FILE *ferr)
{
	int i, j;
	char c;
	for(i=0; isspace(s[i]); ++i);
	for(j=i; !isspace(s[j]) && s[j] != '\0'; ++j);
	myfree(*aname);
	c = s[j]; s[j] = '\0'; *aname = mystrdup(s + i); s[j] = c;
	if(flag == NULL)
	{
	  if(sscanf(s + j, "%lf %lf %lf", x, x+1, x+2)!=3) {
	    if(ferr) fprintf(ferr, "Gread: error near |%s| (%s)\n", s, s+j);
	    return 1;
	  }
	} else {
	  if(sscanf(s + j, "%lf %lf %lf %x", x, x+1, x+2, flag)!=4) {
	    if(ferr) fprintf(ferr, "Gread: error near |%s| (%s)\n", s, s+j);
	    return 1;
	  }
	}

	return 0;
}

static int is_blank_line(char *s)
{
	while((*s) != '\0') {
	  if(!isspace(*s)) return 0;
	  ++s;
	}
	return 1;
}

Geometry *GreadWithError(FILE *f, FILE *ferr)
{
  Geometry *g = NULL;
  int i, j, n;
  double xbuf[3];
  struct Rbuf *rb = NULL;

  do { /* skip blank lines in input */
    if((rb = fgets_rbuf(rb, f)) == NULL) {
      dmd_errno = DMD_INVALID_INPUT;
      goto bail;
    }
  } while(is_blank_line(rb->buf));

  if(sscanf(rb->buf, "%d", &n) == 1 && n > 0) {

    /* XYZ format (XMol, University of Minnesota) 
     * All distances are in Angstroms, no periodicity assumed.
     * Masses are determined from the atomic label.
     * This does not support all XMol options.
     * No flags are returned.
     * The geometry is put into a unit box.
     */
    if(fgets_rbuf(rb, f)==NULL)
    {
      if(ferr) fprintf(ferr, MODULE ": cannot read XYZ title\n");
      dmd_errno = DMD_INVALID_INPUT;
      goto bail;
    }
    for(i=0; rb->buf[i]!='\0'; ++i)if(rb->buf[i]=='\n')rb->buf[i]='\0';
    g = Gmalloc(n, rb->buf);
    g->period = 0;

    for(i=0; i<n; ++i) {
      if(fgets_rbuf(rb, f)==NULL)
      {
      	if(ferr) fprintf(ferr, MODULE ": cannot read XYZ coordinate %d\n", i);
      	Gfree(g); 
	dmd_errno = DMD_INVALID_INPUT;
      	g = NULL;
      	goto bail;
      }
      if(gread_x(rb->buf, g->aname + i, xbuf, NULL, ferr)) goto bail;
      for(j=0; j<3; ++j) xbuf[j] /= ANGSTROM;
      mmul3(g->boxi, xbuf, g->x + 3*i);
    }

  } else if(strstr(rb->buf, "general") != NULL) {

    /* General input, full information about geometry.
     * Read data directly: 
     *   o  optionally, extra keywords on first line: "lattice"
     *   o  periodicity and lattice constant "a"
     *   o  arbitrary box transformation (box = input * a)
     *   o  cartesian coordinates, in box or physical space
     *      depending upon "lattice" keyword.
     */

    int px, py, pz, lattice=0, flags=0, incomplete=0;
    double x[3], det, a;

    /* Check keywords
     */
    if(strstr(rb->buf, "lattice") != NULL)lattice = 1;
    if(strstr(rb->buf, "flags") != NULL)flags = 1;
    if(strstr(rb->buf, "incomplete") != NULL)incomplete = 1;
    if(fgets_rbuf(rb, f)==NULL)
    {
    	if(ferr) fprintf(ferr, MODULE ": cannot read geometry title line\n");
    	goto bail;
    }
    for(i=0; rb->buf[i]!='\0'; ++i)if(rb->buf[i]=='\n')rb->buf[i]='\0';
    if(fscanf(f, "%d (%d,%d,%d) %lf", &n, &px, &py, &pz, &a)!=5)
    {
    	if(ferr) fprintf(ferr, MODULE ": cannot grok geometry info line\n");
    	goto bail;
    }
    g = Gmalloc(n, rb->buf);
    if(flags) g->flags = mymalloc(n * sizeof(int));
    g->period = px | (py << 1) | (pz << 2);
    if(incomplete) g->period |= GP_INCOMPLETE;
    for(i=0; i<9; ++i)
      if(fscanf(f, "%lf", g->box+i)!=1)
      {
        if(ferr) fprintf(ferr, MODULE ": cannot read box member %d\n", i);
        Gfree(g);
	dmd_errno = DMD_INVALID_INPUT;
        g = NULL;
        goto bail;
      }
    for(i=0; i<9; ++i) g->box[i] *= a;
    matinv3(g->boxi, g->box, &det);
    if(ABS(det) < 1.e-10) {
      Gfree(g);
      dmd_errno = DMD_INVALID_INPUT;
      g = NULL;
      goto bail;
    }
    for(i=0; i<n; ++i) {
      do {
        if(fgets_rbuf(rb, f)==NULL)
        {
	  if(ferr) 
	    fprintf(ferr, MODULE ": cannot read general coordinate %d\n", i);
	  Gfree(g);
	  dmd_errno = DMD_INVALID_INPUT;
	  g = NULL;
	  goto bail;
	}
	for(j=0; isspace(rb->buf[j]); ++j);
      } while(rb->buf[j] == '\0');
      if(gread_x(rb->buf, g->aname + i, x, (flags ? g->flags + i : NULL), ferr))
	goto bail;
      if(lattice) {
	for(j=0; j<3; ++j)(g->x + 3*i)[j] = x[j];
      } else {
	(void) mmul3(g->boxi, x, g->x+3*i);
      }
    }

  } else
    if(ferr) fprintf(ferr, MODULE ": unrecognized geometry \"%s\"\n", rb->buf);

bail:
  delete_rbuf(rb);
  return g;
}

Geometry *Gread(FILE *f)
{
	return GreadWithError(f, stderr);
}

