/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gsetbox.c,v 1.6 1996/02/02 15:41:09 deaven Exp $
 * $Author: deaven $
 */

#include "geometry.h"
#define TINY 1.e-6

/* Set the box for a geometry to be equal to the given box.  If
 * the geometry is currently periodic, it will still be marked so;
 * however the physical meaning of the translation symmetry will
 * be different!
 *
 * All of the box coordinates are updated to reflect the change.
 *
 * David Deaven <deaven@ishmael.ameslab.gov>
 */
Geometry *Gsetbox(Geometry *g, double *newbox)
{
	int i;
	double a, x[3];
	static double tinybox[9] =
	  {TINY, 0.0, 0.0,  0.0, TINY, 0.0,  0.0, 0.0, TINY};

	matinv3(g->boxi, newbox, &a);
	if (a < 0.0) a *= -1.0;
	if(a < (0.1 * TINY * TINY * TINY))
	{
	    fprintf(stderr, "Gsetbox: zero volume box (%g)\n", a);
	    return Gsetbox(g, tinybox); /* XXX useful behavior? */
	}
	for(i=0; i<g->n; ++i)
	{
	  mmul3(g->box, g->x+3*i, x);
	  mmul3(g->boxi, x, g->x+3*i);
	}
	for(i=0; i<9; ++i)
	  g->box[i] = newbox[i];

	return g;
}

/* Find the smallest scaling of the geometry's box that will completely 
 * enclose the geometry, so that all points range from -0.5 to +0.5.
 * If a direction is periodic, or the box coordinates in a given
 * direction are all zero, that direction is excluded from adjustment
 * and the box is unchanged from the passed geometry's box.
 * Incomplete geometries are processed exactly the same as are
 * complete geometries.
 */
double *Genclosebox(Geometry *g, double box[3*3])
{
	int i, j;
	double xmax[3], x;

	for(i=0; i<9; ++i) box[i] = g->box[i];

	/* Find the extent of each box coordinate.
	 */
	for(i=0; i<g->n; ++i) {
	  for(j=0; j<3; ++j) {
	    x = g->x[3*i + j];
	    if(x < 0.0) x = -x;
	   if(i==0 || x > xmax[j]) xmax[j] = x;
	  }
	}

	/* Scale the new box to make the extent equal to unity in
	 * each non-periodic direction.
	 * If the coordinates have NO extent in a direction, then
	 * leave the current box size unchanged in that direction.
	 */
	for(i=0; i<3; ++i) {
	  if(!(g->period & GP_PDIR(i)) && xmax[i] > 1.e-10)
	  {
	    for(j=0; j<3; ++j) box[j*3 + i] *= 2.0 * xmax[i];
	  }
	}

	return box;
}

/* Scale the box so that non-periodic directions just fit
 * around the geometry, translating the coordinates if
 * necessary.
 *
 * Do not change box dimensions in periodic directions.
 */
Geometry *GsetboxStandard(Geometry *g)
{
    int i, j;
    double newbox[9];
    double xmin[3], xmax[3], *x, a;

    for(i=0; i<9; ++i) newbox[i] = g->box[i];

    /* Find the extent of each box coordinate.
     */
    for(i=0; i<g->n; ++i) {
      x = g->x + 3 * i;
      for(j=0; j<3; ++j)
      {
	 if(i==0 || x[j] < xmin[j]) xmin[j] = x[j];
	 if(i==0 || x[j] > xmax[j]) xmax[j] = x[j];
      }
    }

    /* Scale the new box to make the extent equal to unity in
     * each non-periodic direction, and translate to make the
     * coordinates fall in the range [-0.5,0.5].
     * If the coordinates have NO extent in a direction, then
     * leave the current box size unchanged in that direction.
     */
    for(i=0; i<3; ++i)
      if(!((g->period >> i) & 1))
      {
	a = xmax[i] - xmin[i];
	if(a > 1.e-10) {
	  for(j=0; j<3; ++j) newbox[j*3 + i] *= a;
	}
	for(j=0; j<(g->n); ++j)
	  g->x[3*j + i] -= (xmin[i] + 0.5 * a);
      }

    return Gsetbox(g, newbox);
}

/* Set the box to a rectangular box that just fits around the 
 * coordinates, translating the coordinates if necessary.
 * This eliminates any periodicity if the box is
 * not rectangular to begin with.
 */
Geometry *GsetboxRect(Geometry *g)
{
    int i, j, isrect = 1;
    double newbox[9];

    for(i=0; i<3; ++i) {
      for(j=0; j<3; ++j) {
	if(i != j && g->box[i + 3*j] != 0.0) isrect = 0;
      }
    }

    if(isrect) {
      for(i=0; i<9; ++i) newbox[i] = g->box[i];
    } else {
      g->period &= (~GP_PERIODIC);
      for(i=0; i<9; ++i) newbox[i] = 0.0;
      newbox[0] = newbox[4] = newbox[8] = 1.0;
    }

    return GsetboxStandard(Gsetbox(g, newbox));
}

/* Set the box to a cubical box that just fits around the
 * geometry, translating the coordinates if nescessary.
 * Eliminates periodiocity.
 */
Geometry *GsetboxCube(Geometry *g)
{
    int i;
    double newbox[9];

    GsetboxRect(g);
    for(i=0; i<9; ++i) newbox[i] = g->box[i];

    if(newbox[4] > newbox[0]) newbox[0] = newbox[4];
    if(newbox[8] > newbox[0]) newbox[0] = newbox[8];
    newbox[4] = newbox[8] = newbox[0];

    return Gsetbox(g, newbox);
}

