/* invupd.c -- invert general complex matrices
 * This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: invupd.c,v 1.4 1996/05/29 21:16:48 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "memory.h"
#include "tools.h"
#include "invupd.h"

/* Given an n by n matrix G (gr/gi), determine the inverse matrix A
 * (ar/ai) and the determinant of G (dr/di).  Use the inverse updating
 * scheme to build the inverse column by column. The calling program
 * must provide phir(*), phii(*), and done(*) to be used as scratch
 * space.
 *
 * Created JUL 1992  From old code specific to Wick matrices.
 *                   Author: David Deaven (DMD)
 *
 * Modified FEB 1992 To perform selective column updates reliably.
 *                   Fixes bug in skip code, and makes zeroes on
 *                   the diagonal OK (DMD).
 */
#define TINY 1.e-20
int matinv(double gr[], double gi[], int n, 
	double ar[], double aip[], double *dr, double *di)
{
    int i, j, k, s;
    double q, r, *ai;
    int ndone, nskip;
    double qi, qr;

    double *phir = dmalloc(n), *phii = dmalloc(n);
    int *done = imalloc(n);
    if(aip == NULL) ai = dmalloc(n*n);
    else ai = aip;

    /* I. Make the identity matrix, det=1. */

    for (i = 0; i < n; ++i) {
	for (j = 0; j < n; ++j) {
	    ar[i + j*n] = (i==j ? 1.0 : 0.0);
	    ai[i + j*n] = 0.0;
	}
	done[i] = 0;
    }
    *dr = 1.0;
    *di = 0.0;

    /* II. Build the desired matrix column by column. 
     *     Skip columns which when updated produce a singular matrix,
     *     singular is defined by det A < TINY.
     *     If no column may be updated without producing a singular 
     *     matrix, then try adding (1,...,1) to a column and _then_
     *     checking for updatable columns. Columns with (1,...,1) added
     *     are marked by done(k)=-1.
     */

    k = 0;
    nskip = ndone = 0;
    while(ndone < n) {
	while(done[k] == 1) if (++k >= n) k = 0;
	for (j = 0; j < n; ++j) {
	    phir[j] = gr[j + k * n];
	    phii[j] = (gi == NULL ? 0.0 : gi[j + k * n]);
	    if (k == j) phir[j] -= 1.0;
	}
	ratest(ar, ai, n, phir, phii, k, &qr, &qi, &q);
	if (q < TINY) {
	    if (++nskip > 2*n) {
		fprintf(stderr, "matinv(): WARNING: cannot update all columns\n");
		*dr = *di = 0.0;
		s = 0;
		goto bail;
	    }
	    if (nskip >= n) {
		for (j = 0; j < n; ++j) {
		    phir[j] = 1.0;
		    phii[j] = 0.0;
		}
		phir[k] = 0.0;
		ratest(ar, ai, n, phir, phii, k, &qr, &qi, &q);
		if (q > TINY) {
		    invupd(ar, ai, n, phir, phii, k, qr, qi);
		    r = (*dr) * qr - (*di) * qi;
		    *di = (*dr) * qi + (*di) * qr;
		    *dr = r;
		    done[k] = -1;
		}
	    }
	    if (++k >= n) k = 0;
	} else {
	    invupd(ar, ai, n, phir, phii, k, qr, qi);
	    r = (*dr) * qr - (*di) * qi;
	    *di = (*dr) * qi + (*di) * qr;
	    *dr = r;
	    ++ndone;
	    if (done[k] == -1) {
		for (j = 0; j < n; ++j) {
		    phir[j] = -1.0;
		    phii[j] =  0.0;
		}
		phir[k] = 0.0;
		ratest(ar, ai, n, phir, phii, k, &qr, &qi, &q);
		if (q < TINY && ndone < n) {
		  dmd_fatal_error("matinv(): "
		    "I painted myself into a corner!\n");
		} else {
		    if (q >= TINY) {
			invupd(ar, ai, n, phir, phii, k, qr, qi);
			r = (*dr) * qr - (*di) * qi;
			*di = (*dr) * qi + (*di) * qr;
			*dr = r;
		    } else {
			*dr = *di = 0.0;
		    }
		}
	    }
	    done[k] = 1;
	}
    }
    s = 0;
bail:
    myfree(phir); myfree(phii); myfree(done);
    if(aip == NULL) myfree(ai);
    return s;
}

/* Given the inverse of an nxn matrix A stored in ar/ai, and a vector
 * phir/phii which is to be added to the m'th column of A, and the ratio
 * of the new determinant to the old determinant qr/qi, update ar and
 * ai to reflect the new inverse.
 *
 * David Deaven April 1991
 */
void invupd(double ar[], double ai[], int n, 
	double phir[], double phii[], int m, 
	double qr, double qi)
{
    double a;
    int i, j;
    double q2, bi, ci, di, br, cr, dr;

    /* I. Find d=1/q */
    q2 = qr*qr + qi*qi;
    dr = qr / q2;
    di = - qi / q2;

    /* II. Update the inverse. */
    for (i = 0; i < n; ++i) {
	if (i != m) {
	    br = bi = 0.0;
	    for (j = 0; j < n; ++j) {
		br += ar[i + j*n] * phir[j] - ai[i + j*n] * phii[j];
		bi += ai[i + j*n] * phir[j] + ar[i + j*n] * phii[j];
	    }
	    cr = br * dr - bi * di;
	    ci = br * di + bi * dr;
	    for (j = 0; j < n; ++j) {
		ar[i + j*n] += - cr * ar[m + j*n] + ci * ai[m + j*n];
		ai[i + j*n] += - cr * ai[m + j*n] - ci * ar[m + j*n];
	    }
	}
    }
    for (j = 0; j < n; ++j) {
	a = dr * ar[m + j*n] - di * ai[m + j*n];
	ai[m + j*n] = dr * ai[m + j*n] + di * ar[m + j*n];
	ar[m + j*n] = a;
    }

}

/* Given the inverse of an nxn matrix A stored in ar/ai, and a vector
 * phir/phii which is to be added to the m'th column of A, find the ratio 
 * of the new to old determinants, and return it in qr/qi. Also return
 * the magnitude of the ratio in q.
 */
void ratest(double ar[], double ai[], int n, 
	double phir[], double phii[], int m, 
	double *qr, double *qi, double *q)
{
    int i;

    *qr = 1.0;
    *qi = 0.0;
    for (i = 0; i < n; ++i) {
	*qr += ar[m + i*n] * phir[i] - ai[m + i*n] * phii[i];
	*qi += ai[m + i*n] * phir[i] + ar[m + i*n] * phii[i];
    }
    *q = sqrt((*qr) * (*qr) + (*qi) * (*qi));
}
