/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: memory2.c,v 1.11 1996/08/08 19:48:17 deaven Exp $
 * $Author: deaven $
 */

#ifdef TRACKMEM
#undef TRACKMEM
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "memory.h"
#include "tools.h"

/* Offenses at a given file and line are recorded for the summary.
 * After a threshold number of warnings, they are no longer flagged
 * dynamically.
 */
#define MAX_OFFENSE_WARNINGS 3

/* If this is non-zero, it is a size in bytes of the smallest reported
 * change in the total allocated memory.  If it is zero, no reports
 * about changes in total reported memory are reported.
 */
unsigned int memory_report_chunk = 0;

struct offense {
	char *file;
	int line;
	int warnings;
	struct offense *next;
};

struct block {
	void *loc;	/* starting location in memory */
	size_t size;	/* size of allocated memory */
	char *file;	/* source file in which block was last allocated */
	int line;	/* source line number in which last allocated */
	struct block *next;	/* next block */
};

static struct offense *offense = NULL;
static struct block *base = NULL;
static size_t total = 0;

static struct offense *new_offense(char *file, int line)
{
	struct offense *o = malloc(sizeof(*o));

	if(o == NULL)
	  dmd_fatal_error("allocating trackmem offense record");
	o->file = file;
	o->line = line;
	o->warnings = 0;
	o->next = NULL;

	return o;
}

static void fatal(char *op, size_t size, char *file, int line)
{
	char buf[FILENAME_MAX + 256];
	sprintf(buf, "pid %d: cannot %s (size = %d) file `%s' line %d\n",
	  (int) getpid(), op, (int) size, file, line);
	mymemreport_x(stderr);
	dmd_fatal_error(buf);
}

/* Format a memory size using the correct units.
 */
static char *typemem(char *s, long m)
{
	if(m < 1024) {
	  sprintf(s, "%db", (int) m);
	} else {
	  char *unit;
	  double d = m;
	  d /= 1024.0; unit = "kb";
	  if(d > 512.0) { d /= 1024.0; unit = "Mb"; }
	  if(d > 512.0) { d /= 1024.0; unit = "Gb"; } /* ha */
	  if(d > 512.0) { d /= 1024.0; unit = "Tb"; } /* HA HA */
	  sprintf(s, "%.2f%s", d, unit);
	}
	return s;
}

/* Report significant changes in allocated memory.
 */
static void check(char *s, size_t size, char *file, int line)
{
	static size_t warn = 0;
	long del = total - warn;
	if(del < 0) del = -del;
	if (memory_report_chunk && del > memory_report_chunk)
	{
	  char buffer[128], *next = buffer;
	  typemem(next, del);
	  while(*next != '\0') ++next;
	  *(next++) = '/';
	  typemem(next, (long) total);
	  fprintf(stderr, "[%s(%ld) %s `%s' line %d]\n",
	    s, (long) size, buffer, file, line);
	  warn = total;
	}
}

static struct block *add_block(void *loc, size_t size, char *file, int line)
{
	struct block *new;

	if((new = malloc(sizeof(struct block))) == NULL) {
	  dmd_fatal_error("memory2: cannot allocate info block\n");
	}

	new->next = base;
	new->loc = loc;
	new->size = size;
	new->file = file;
	new->line = line;
	base = new;

	return new;
}

/* Report an incorrect memory allocation operation.
 *
 * This is always called if incorrect allocation usage occurs, so
 * it may be used as a breakpoint in a debugger.
 */
static void trackmem(char *s, void *ptr, char *file, int line)
{
	struct offense *o;

	for(o = offense; o != NULL; o = o->next) {
	  if(line == o->line && strcmp(file, o->file) == 0) break;
	}

	if(o == NULL) {
	  o = new_offense(file, line);
	  o->next = offense;
	  offense = o;
	}

	o->warnings++;

	if(o->warnings <= MAX_OFFENSE_WARNINGS) {
	  fprintf(stderr, "trackmem(): %s non-existant block at %p: "
	    "file `%s' line %d (%d%s)\n", s, ptr, file, line,
	    o->warnings, o->warnings == MAX_OFFENSE_WARNINGS ?
	      ", further warnings suppressed" : "");
	}
}

/* Re-allocate a block, or allocate it if NULL pointer is passed.
 */
void *myrealloc_x(void *ptr, size_t size, char *file, int line)
{
	void *r;
	struct block *b;

	if(ptr == NULL) return mymalloc_x(size, file, line);

	for(b = base; b != NULL; b = b->next) if (b->loc == ptr) break;
	
	if (b == NULL) {
	  trackmem("reallocate", ptr, file, line);
	  b = add_block(ptr, 0, file, line);
	}

	if(size == 0) {
	  myfree_x(ptr, file, line);
	  return NULL;
	}

	if((r = realloc(ptr, size)) == NULL)
	  fatal("myrealloc", size, file, line);
	
	total += size - b->size;
	b->loc = r;
	b->size = size;
	b->file = file;
	b->line = line;
	if(memory_report_chunk) check("realloc", size, file, line);
       
	return r;
}

/* Allocate a block.
 */
void *mymalloc_x(size_t size, char *file, int line)
{
	void *r;

	if(size == 0) return NULL;

	r = malloc(size);

	if(r == NULL) fatal("mymalloc", size, file, line);

	(void) add_block(r, size, file, line);
	total += size;
	if(memory_report_chunk) check("alloc", size, file, line);

	return r;
}

/* Free a memory block 
 */
void myfree_x(void *ptr, char *file, int line)
{
	struct block *b, *bl;
	if(ptr == NULL) return;
	
	for(bl = NULL, b = base; b != NULL; b = b->next) {
	  if (b->loc == ptr) break;
	  bl = b;
	}
	if (b == NULL) {
	  trackmem("free", ptr, file, line);
	} else {
	  total -= b->size;
	  if(bl == NULL)		/* b is the current base */
	    base = b->next;
	  else			/* bl precedes b */
	    bl->next = b->next;
	  if(memory_report_chunk) check("free", b->size, file, line);
	  free(b);
	}

	(void) free(ptr);
}

/* A replacement for strdup(), which is not required by ANSI
 * and so is not always present.
 */
char *mystrdup_x(char *s, char *file, int line)
{
	int n = 0;
	char *copy;
	if(s==NULL) return NULL;
	while(s[n] != '\0') ++n;
	copy = mymalloc_x((n+1) * sizeof(char), file, line);
	for(; n>=0; --n) copy[n] = s[n];
	return copy;
}

/* Report on memory usage.
 */
void mymemreport_x(FILE *f)
{
	struct block *b;
	char buf[128];
	fprintf(f, "Memory allocation map: (total = %s)\n", 
	  typemem(buf, (long) total));
	fprintf(f, "%s\t%10s %s\n", "loc", "size", "file:line");
	for(b = base; b != NULL; b = b->next)
	{
		fprintf(f, "%p\t%10s %s:%d\n",
		  b->loc, typemem(buf, (long) b->size), b->file, b->line);
	}
}

/* This is included so that a debugger like gdb can call it to track
 * the current state of affairs during debugging.
 */
void debug_mem()
{
	mymemreport_x(stdout);
}
