/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: optdfp.c,v 1.7 1996/03/20 18:07:22 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "tools.h"
#include "memory.h"

#define STEP0	0.1	/* initial step ratio */
#define MAXLOOP	3	/* maximum loops through line minimization */
#define DOTMIN	0.1	/* minimum dot product for direction before restart */
#define ZEPS	1.e-20	/* too small for vectors norms, signals restart */

#define MYMIN(x,y) (x < y ? (x) : (y))
#define MYMAX(x,y) (x > y ? (x) : (y))

static void ident_matrix(double *a, int n)
{
	int i;
	for(i=0; i<(n * n); ++i) a[i] = 0.0;
	for(i=0; i<n; ++i) a[i*n + i] = 1.0;
}

/* Davidon-Fletcher-Powell optimization technique on coordinates
 *
 * Input:
 *   x[] = coordinates
 *   n = number of coordinates
 *   frceps = maximum force smaller than this, and...
 *   eps = ...relative change in energy smaller than this = convergence
 *   itmax = maximum number of iterations allowed
 *   irmax = maximum number of restarts allowed
 *   hmat[] = initial guess for inverse Hessian (unit matrix OK)
 *     or NULL to allocate, start with unit, and deallocate before
 *     returning.
 *   callback(x,f,e,n,resources) = function to be called after each line
 *     minimization -- must NOT modify the positions or forces!
 *     If callback() is NULL no action is taken.
 *   fun(x,f,e,n,resources) = function used to compute forces and energy.
 *     It can return non-zero to indicate invalid parameters in x[].
 *   fun_resources = data passed to fun() and callback().
 *   fout = optional file for information
 *
 * Output:
 *   f[] = forces on coordinates at converged point.
 *   hmat[] = inverse of approximate Hessian matrix.
 *
 * David Deaven 15 February 1994 <deaven@ishmael.ameslab.gov>
 * 16 August 1994: add "extra try" to parabolic loop code.
 * 25 January 1996: add callback function
 *
 */
int optdfp_engine(double x[], double f[], double *ep, int n,
  double frceps, double eps, double step0, double dotmin, int maxloop,
  int use_bfgs, int itmax, int irmax, double *hmat_passed, 
  void *fun_resources,
  void (*callback)(double *, double *, double *, int, void *),
  int (*fun)(double *, double *, double *, int, void *), FILE *fout)
{
  int i, j, iter, nstart, para_nstart, r, nscan;
  double *sdir, *lastf, *bfgs = NULL, *hdg;
  double dx, step, para_step, eptol, a, b, s, fnorm, snorm, dot;
  double va, vb, ya, yb, vc, t, frcmax;
  double theta, *hmat;
  double dx_dg, dx_dg_i, dg_h_dg_i, dg_h_dg;

  sdir = dmalloc(n);
  lastf = dmalloc(n);
  hdg = dmalloc(n);
  if(use_bfgs) bfgs = dmalloc(n);

  if(hmat_passed == NULL)
  {
    hmat = dmalloc(n * n);
    ident_matrix(hmat, n);
  }
  else
    hmat = hmat_passed;

  iter = nstart = 0;
  if(fout) fprintf(fout, "%3s %8s %12s %12s %12s %12s\n",
       "it", "theta", "Fmax", "dE/|E|", "stepsize", "E");

  /* Use a small step size to begin with.
   */
  step = step0;
  nscan = 1;
  do {
    if(++iter > itmax) {
      if(fout) fprintf(fout, "optdfp: too many iterations: %d\n", iter);
      r=-1; goto bailout;
    }

    /* Get local gradient and adjust step size, depending upon
     * the last number of 1d scans required to bracket minimum.
     */
    if((r=(*fun)(x, f, ep, n, fun_resources))) {
      if(fout)
	fprintf(fout, "optdfp: function failed starting iteration %d\n", iter);
      goto bailout;
    }
    if(nscan==0) step *= 0.618;
    step = fabs(step);
    eptol = *ep;

    /* Perform callback if requested.
     */
    if(callback != NULL) callback(x, f, ep, n, fun_resources);

    /* Deflect to search direction using accumulated inverse Hessian,
     * and get normalization factor, dot product.
     * Also save current positions and forces for later adjustment
     * of the inverse Hessian.
     * We maintain an angle theta between the step direction and the
     * local gradient, to check stability of the algorithm.
     */
    do {
      fnorm = snorm = dot = 0.0;
      for(i=0; i<n; ++i)
      {
	lastf[i] = f[i];
	sdir[i] = 0.0;
	for(j=0; j<n; ++j) sdir[i] += hmat[i+j*n] * f[j];
	fnorm += f[i] * f[i];
	snorm += sdir[i] * sdir[i];
	dot += f[i] * sdir[i];
      }

      fnorm = sqrt(fnorm);
      snorm = sqrt(snorm);

      if(fnorm <= ZEPS)
	{r=0; goto bailout;} /* almost impossible: we're done */

      /* Several conditions may cause a "restart" of the approximate
       * inverse Hessian matrix.  If this happens, the identity matrix
       * is used as the new guess.
       */
      if(snorm <= ZEPS || dot < (dotmin * snorm * fnorm))
      {
	if(nstart++ > irmax) {
	  if(fout) fprintf(fout, "optdfp: too many restarts: %d\n", nstart);
	  r=-1; goto bailout;
	}
	else
	  if(fout) fprintf(fout,
	    "optdfp: restart %d in Hessian (dot=%.4g S=%.4g F=%.4g)\n",
	    nstart, dot / (snorm * fnorm), snorm, fnorm);
	ident_matrix(hmat, n);
	step = step0;
	nscan = 1;
      } else {
        break;
      }
    } while(1);
    theta = (180.0 / M_PI) * acos(MYMIN(1.0, dot / (snorm * fnorm)));

    for(i=0; i<n; ++i) sdir[i] *= (1.0 / snorm);

    /* Do a 1d search along the search direction, using Davidon search
     * algorithm (parabolic interpolation).
     * If the function bombs out, or the increase is huge, try to recover 
     * gracefully by reducing the step size by a large factor.
     * The final updated point x[] will be lastx[] + dx * sdir[] (I
     * do not actually store lastx[]).
     */
    dx = 0.0;
    para_step = step;	/* for this search loop only */
    para_nstart = 0;
    do {
      /* Top of search loop, find {ya,yb} bracketing minimum */
      yb = *ep;
      vb = dot;
      nscan = 0;
      do {
	ya = yb;
	va = vb;

	for(i=0; i<n; ++i) x[i] += step * sdir[i];
	dx += step;
	if((r=(*fun)(x, f, ep, n, fun_resources)) ||
	   (*ep - ya) > fabs(10.0 * ya) )
	{
	  for(i=0; i<n; ++i) x[i] -= 0.9 * step * sdir[i];
	  dx -= 0.9 * step;
	  step *= 0.1;
	  if((r=(*fun)(x, f, ep, n, fun_resources))) {
	    if(fout)
	      fprintf(fout, "optdfp: bracketing recovery failed: step=%g\n",
		step);
	    goto bailout;
	  }
	}

	yb = *ep;
	for(vb=0.0, i=0; i<n; ++i) vb += f[i] * sdir[i] * snorm;
	if(vb > 0.0 && yb < ya)
	{
	  ++nscan;
	  step *= 1.618;
	}

      } while(vb > 0.0 && yb < ya);

      /* Parabolic interpolation to minimum.
       */
      t=0.0;
      for(j=maxloop; j>0; --j)
      {
	a = 3.0*(ya-yb)*snorm/step-va-vb;
	if((b = a*a-va*vb) < ZEPS) break; /* inconsistant or jagged function */
	b = sqrt(b);
	s = step*(-vb+b-a)/(-vb+va+2.0*b);
	if(fabs(s/step) > 2.0) {
	  fprintf(stderr, "optdfp: long jump averted: s/step=%e\n", s/step);
	  s *= 2.0 * fabs(step/s);
	}
	for(i=0; i<n; ++i) x[i] += (t - s) * sdir[i];
	dx += (t - s);
	if((r=(*fun)(x, f, ep, n, fun_resources))) {
	  if(fout)
	    fprintf(fout,
	      "optdfp: function failed in parabolic step: s/step=%e",
	      s / step);
	  goto bailout;
	}
	if((*ep) < ya && (*ep) < yb) break; /* we hit a relative minimum OK */
	for(vc=0.0, i=0; i<n; ++i)
	  vc += f[i] * sdir[i] * snorm; /* damage control */
	if(vc < 0.0)
	{
	  yb = *ep;
	  vb = vc;
	  step -= s;
	  t = 0.0;
	}
	else
	{
	  ya = *ep;
	  va = vc;
	  step = t = s;
	}
      }

      /* If we cannot find the minimum, best to reset the step size
       * and try again.  The new step size is obtained by dividing
       * down the initial step size at the beginning of the search
       * loop by a power of the number of loop-driven restarts.
       *
       * We count this as a restart, though it is of a different nature
       * than the one above.
       */
      if(j == 0) /* messed up somehow */
      {
	++para_nstart;
	if(nstart++ > irmax)
	{
	  if(fout) fprintf(fout,
	    "optdfp: too many restarts in para: %d\n", nstart-1);
	  r=-1; goto bailout;
	}
	else
	  if(fout) fprintf(fout,
	    "optdfp: restart %d in parabolic interpolation\n", nstart);

	step = para_step;
	for(i=0; i<para_nstart; ++i) step *= (1.0 / 3.0);
      }
      else
        break;
    } while(1); /* 1d search loop */

    eptol = (*ep-eptol)/fabs(*ep);
    frcmax = 0.0;
    for(i=0; i<n; ++i)frcmax = MYMAX(frcmax, fabs(f[i]));

    /* Update the appoximate inverse Hessian matrix.
     */
    for(i=0; i<n; ++i) {
      hdg[i] = 0.0;
      for(j=0; j<n; ++j)
	hdg[i] -= hmat[i*n + j] * (f[j]-lastf[j]);
    }
    dx_dg = dg_h_dg = 0.0;
    for(i=0; i<n; ++i) {
      dx_dg -= dx * sdir[i] * (f[i]-lastf[i]);
      dg_h_dg -= hdg[i] * (f[i]-lastf[i]);
    }

    if(dx_dg !=0.0 && dg_h_dg != 0.0)
    {
      dx_dg_i = 1.0 / dx_dg;
      dg_h_dg_i = 1.0 / dg_h_dg;
      if(use_bfgs)
	for(i=0; i<n; ++i)
	  bfgs[i] = dx_dg_i * dx * sdir[i] - dg_h_dg_i * hdg[i];
      for(i=0; i<n; ++i)
	for(j=0; j<n; ++j) {
	  hmat[i*n + j] +=
#if 0
	    /* This is the expression that is found in Numerical Recipes,
	     * but the following one works much better. XXX
	     */
	    dx_dg_i * dx * sdir[i] * dx * sdir[j]
#else
	    dx_dg_i * dx * sdir[i] * snorm * sdir[j]
#endif
	  - dg_h_dg_i * hdg[i] * hdg[j]
	  + (use_bfgs ? dg_h_dg * bfgs[i] * bfgs[j] : 0.0);
	}
    }

    /* Check convergence, and do another iteration unless we are done
     */
    if(fout) fprintf(fout, "%3d %8.4f %12.4e %12.4e %12.4e %12.4e\n",
       iter, theta, frcmax, eptol, step, *ep);

  } while(frcmax > frceps || fabs(eptol) > eps);
  r = 0;

bailout:
  myfree(sdir);
  myfree(lastf);
  myfree(hdg);
  myfree(bfgs);
  if(hmat_passed==NULL) myfree(hmat);
  return r;
}

int optdfp(double x[], double f[], double *ep, int n,
  double frceps, double eps,
  int itmax, int irmax, double *hmat_passed, 
  void *fun_resources,
  int (*fun)(double *, double *, double *, int, void *), FILE *fout)
{
	return optdfp_engine(x, f, ep, n, frceps, eps, STEP0,
	  DOTMIN, MAXLOOP, 0, itmax, irmax, hmat_passed,
	  fun_resources, NULL, fun, fout);
}

int optbfgs(double x[], double f[], double *ep, int n,
  double frceps, double eps,
  int itmax, int irmax, double *hmat_passed, 
  void *fun_resources,
  int (*fun)(double *, double *, double *, int, void *), FILE *fout)
{
	return optdfp_engine(x, f, ep, n, frceps, eps, STEP0,
	  DOTMIN, MAXLOOP, 1, itmax, irmax, hmat_passed,
	  fun_resources, NULL, fun, fout);
}

