/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: quaternion.c,v 1.4 1996/07/09 16:36:10 deaven Exp $
 * $Author: deaven $
 */

/* Manipulating quaternions, rotations
 */
#include <math.h>
#include "dmd/memory.h"
#include "dmd/tools.h"

/* Generate a rotational transformation using quaternions.
 * yaw around z, pitch around local y, roll around final x.
 *
 * This is the transpose of Goldstein Eq.(4-67), since we are
 * transforming coordinates instead of axes.
 */
double *QRotationMatrix(double q[4], double t[3*3])
{
	t[0] = q[0]*q[0] + q[1]*q[1] - q[2]*q[2] - q[3]*q[3];
	t[3] = 2.0 * (q[1]*q[2] + q[0]*q[3]);
	t[6] = 2.0 * (q[1]*q[3] - q[0]*q[2]);

	t[1] = 2.0 * (q[1]*q[2] - q[0]*q[3]);
	t[4] = q[0]*q[0] - q[1]*q[1] + q[2]*q[2] - q[3]*q[3];
	t[7] = 2.0 * (q[2]*q[3] + q[0]*q[1]);

	t[2] = 2.0 * (q[1]*q[3] + q[0]*q[2]);
	t[5] = 2.0 * (q[2]*q[3] - q[0]*q[1]);
	t[8] = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3];

	return t;
}

struct context {
	int n;
	double *J;
};

void *QNewIntegrateContext(int n)
{
	struct context *c = NewPointer(c);
	c->n = n;
	c->J = NULL;
	return ((void *) c);
}

void QDeleteIntegrationContext(void *context)
{
	struct context *c = (struct context *) context;
	if(c) {
	  myfree(c->J);
	  myfree(c);
	}
}

/* Given for n rigid rotors:
 *    q[4*n], the orientation of each rotor,
 *    w[3*n], the body frame angular velocity of each rotor,
 *    I[3*n], the principal moments of inertia for each rotor,
 *    t[3*n], the space frame torques on each rotor,
 *    h, a timestep;
 * Advance the orientations q[] and body-frame angular
 * velocities w[] given the torque[], and return the rotational
 * kinetic energy.
 *   
 * David Deaven, January 1996.
 */
double QIntegrateSimpson(void *context,
  double *q, double *w, double *I, double h, double *t)
{
	int i, n = ((struct context *) context)->n;
	double *qi, *wi, *ti, *Ii, dqi[4], dwi[3];
	double b[3*3], tis[3];
	double ek = 0.0;
	static double Iunit[3] = {1.0, 1.0, 1.0};

	for(i=0; i<n; i++)
	{
	  qi = q + 4*i;
	  wi = w + 3*i;
	  ti = t + 3*i;
	  Ii = I ? I + 3*i : Iunit;

	  /* These are equations of motion in the body frame.
	   * The dwi[] are the _body_frame_ angular accelerations.
	   * We have to convert the space frame torques (ti[]) to the
	   * body frame (tis[]) first.
	   */
	  mmul3(QRotationMatrix(qi, b), ti, tis);
	  dwi[0] = (tis[0] + wi[1] * wi[2] * (Ii[1] - Ii[2])) / Ii[0];
	  dwi[1] = (tis[1] + wi[2] * wi[0] * (Ii[2] - Ii[0])) / Ii[1];
	  dwi[2] = (tis[2] + wi[0] * wi[1] * (Ii[0] - Ii[1])) / Ii[2];

	  dqi[0] = -qi[1]*wi[0] - qi[2]*wi[1] - qi[3]*wi[2];
	  dqi[1] =  qi[0]*wi[0] - qi[3]*wi[1] + qi[2]*wi[2];
	  dqi[2] =  qi[3]*wi[0] + qi[0]*wi[1] - qi[1]*wi[2];
	  dqi[3] = -qi[2]*wi[0] + qi[1]*wi[1] + qi[0]*wi[2];

	  qi[0] += h * dqi[0];
	  qi[1] += h * dqi[1];
	  qi[2] += h * dqi[2];
	  qi[3] += h * dqi[3];

	  wi[0] += h * dwi[0];
	  wi[1] += h * dwi[1];
	  wi[2] += h * dwi[2];

	  ek += 0.5 * (Ii[0] * wi[0] * wi[0] +
		       Ii[1] * wi[1] * wi[1] +
		       Ii[2] * wi[2] * wi[2]   );
	}

	return ek;
}

/* Given an integration context and some coordinates, angular
 * velocities, principal moments of inertia, and a timestep h, 
 * advance the quaternion values and the angular velocities given that
 * the torque is q.  The angular velocities will be set to the proper
 * value at the current time, while the coordinates will be set to the
 * proper value at time t0 + h.
 *
 * Use the Lax-Wendroff integration scheme.
 * (See _Numerical_Recipes_ for details.)
 */
double QIntegrate(void *context,
  double *q, double *w, double *I, double h, double *t)
{
	int i, j;
	double *qi, *wi, *ti, *Ii, dqi[4];
	double R[3*3], J0[3], w1[3], q1[4];
	double ek = 0.0, norm;
	struct context *c = (struct context *) context;
	static double Iunit[3] = {1.0, 1.0, 1.0};

	/* For initialization, use the current angular momentum.
	 */
	if(c->J == NULL) {
	  c->J = dmalloc(3 * c->n);
	  for(i=0; i<c->n; i++) {
	    qi = q + 4*i;
	    wi = w + 3*i;
	    Ii = I ? I + 3*i : Iunit;
	    for(j=0; j<3; j++) J0[j] = wi[j] * Ii[j];
	    mmul3(QRotationMatrix(qi, R), J0, c->J + 3*i);
	  }
	}

	/* The formulae steps are as follows.  Let discrete times
	 * indexed by -1, 0, 1, 2 be -h/2, 0, h/2, h relative to
	 * the current time.  ^L means lab frame, ^B means body
	 * frame.  Perform the following:
	 *
	 * 1.  J^L_0 = J^L_-1 + h/2 t^L_0
	 * 2.  w^B_0 = (Rt J^L_0) / I^B
	 * 3.  q_1 = q_0 + h/2 Q_0 w^B_0
	 * 4.  J^L_1 = J^L_-1 + h t^L_0
	 * 5.  w^B_1 = (Rt J^L_1) / I^B
	 * 6.  q_2 = q_0 + h Q_1 w^B_1
	 *
	 * Here Q is the transformation matrix that gives time
	 * derivatives of quaternions when multiplied by the
	 * (0, wx, wy, wz) four-vector angular velocity.  Rt is
	 * the transpose of the coordinate rotation matrix.
	 *
	 * The only quantity that has to be saved is the lab
	 * frame angular momentum at the half-step times.
	 */
	for(i=0; i<c->n; i++)
	{
	  qi = q + 4*i;
	  wi = w + 3*i;
	  ti = t + 3*i;
	  Ii = I ? I + 3*i : Iunit;
	  QRotationMatrix(qi, R);

	  for(j=0; j<3; j++) J0[j] = c->J[3*i + j] + 0.5 * h * ti[j];

	  mtmul3(R, J0, wi);
	  for(j=0; j<3; j++) {
	    wi[j] /= Ii[j];
	    ek += 0.5 * Ii[j] * wi[j] * wi[j];
	  }

	  dqi[0] = -qi[1]*wi[0] - qi[2]*wi[1] - qi[3]*wi[2];
	  dqi[1] =  qi[0]*wi[0] - qi[3]*wi[1] + qi[2]*wi[2];
	  dqi[2] =  qi[3]*wi[0] + qi[0]*wi[1] - qi[1]*wi[2];
	  dqi[3] = -qi[2]*wi[0] + qi[1]*wi[1] + qi[0]*wi[2];

	  for(j=0; j<4; j++) q1[j] = qi[j] + 0.5 * h * dqi[j];

	  for(j=0; j<3; j++) c->J[3*i + j] += h * ti[j];

	  mtmul3(R, c->J + 3*i, w1);
	  for(j=0; j<3; j++) w1[j] /= Ii[j];

	  dqi[0] = -q1[1]*w1[0] - q1[2]*w1[1] - q1[3]*w1[2];
	  dqi[1] =  q1[0]*w1[0] - q1[3]*w1[1] + q1[2]*w1[2];
	  dqi[2] =  q1[3]*w1[0] + q1[0]*w1[1] - q1[1]*w1[2];
	  dqi[3] = -q1[2]*w1[0] + q1[1]*w1[1] + q1[0]*w1[2];

	  norm = 0.0;
	  for(j=0; j<4; j++) {
	    qi[j] += h * dqi[j];
	    norm += qi[j] * qi[j];
	  }
	  norm = 1.0 / sqrt(norm);
	  for(j=0; j<4; j++) qi[j] *= norm;
	}

	return ek;
}

/* Convert a quaternion to YPR orientation angles.
 */
double *QtoYPR(double q[4], double ypr[3])
{
	double R[3*3];
	QRotationMatrix(q, R);
	YPRfromR(R, ypr);
	return ypr;
}
